# (c) 2015-2018 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#
from moleculekit.projections.projection import Projection
import numpy as np
import logging
logger = logging.getLogger(__name__)


class Dihedral:
    """ Class to store atoms defining a dihedral angle.

    Example
    -------
    >>> # Using the helper functions to construct a dihedral object
    >>> d1 = Dihedral.phi(mol, 5, 6, segid='P0')
    >>> d2 = Dihedral.chi1(mol, 12, segid='P0') # Defining segid
    >>> d3 = Dihedral.chi1(mol, 38, chain='X') # Defining chain
    >>> # Manual construction
    >>> atom1 = {'name': 'N', 'resid': 5, 'segid': 'P'}
    >>> atom2 = {'name': 'CA', 'resid': 3, 'segid': 'P', 'chain': 'A', 'insertion': 'B'}
    >>> atom3 = {'name': 'C', 'resid': 46, 'segid': 'P', 'chain': 'X'}
    >>> atom4 = {'name': 'O', 'resid': 2, 'segid': 'P'}
    >>> d = Dihedral(atom1, atom2, atom3, atom4)
    """
    def __init__(self, atom1, atom2, atom3, atom4, dihedraltype=None, check_valid=True):
        if check_valid:
            valid_keys = ('name', 'resid', 'segid', 'insertion', 'chain')
            default_values = {'name': '', 'resid': 0, 'segid': '', 'insertion': '', 'chain': ''}
            for i, a in enumerate([atom1, atom2, atom3, atom4]):
                for k in a.keys():
                    if k not in valid_keys:
                        raise RuntimeError('Dictionary key can\'t be "{}". Valid keys are: {}'.format(k, valid_keys))
                for k in np.setdiff1d(valid_keys, tuple(a.keys())):
                    print(np.setdiff1d(valid_keys, tuple(a.keys())))
                    a[k] = default_values[k]
                #frames.append(pd.DataFrame(a, index=[i]))
        #self.atoms = pd.concat(frames)
        self.atoms = [atom1, atom2, atom3, atom4]
        self.dihedraltype = dihedraltype

    def __str__(self):
        descr = ''
        if self.dihedraltype is not None:
            descr = '"{}" dihedral angle including atoms:\n'.format(self.dihedraltype)
        descr += 'name\tresid\tinsertion\tchain\tsegid\n'
        for a in self.atoms:
            descr += '{}\t{}\t{}\t{}\t{}\n'.format(a['name'], a['resid'], a['insertion'], a['chain'], a['segid'])
        return descr

    def __repr__(self):
        return self.__str__()

    def _explanation(self):
        descr = ''
        if self.dihedraltype is not None:
            descr += self.dihedraltype + ' '
        resids = []
        insertions = []
        for a in self.atoms:
            resids.append(a['resid'])
            insertions.append(a['insertion'])
        uqresids = list(np.unique(resids))
        uqinsertions = list(np.unique(insertions))

        descr += 'dihedral of Resid/Insertion/Chain/Segid/ {}/{}/{}/{}'.format(','.join(map(str, uqresids)), ','.join(uqinsertions), self.atoms[0]['chain'], self.atoms[0]['segid'])
        return descr

    @staticmethod
    def dihedralsToIndexes(mol, dihedrals, sel='all'):
        """ Converts dihedral objects to atom indexes of a given Molecule
        
        Parameters
        ----------
        mol : :class:`Molecule <moleculekit.molecule.Molecule>` object
            A Molecule object from which to obtain atom information
        dihedrals : list
            A single dihedral or a list of Dihedral objects
        sel : str
            Atom selection string to restrict the application of the selections.
            See more `here <http://www.ks.uiuc.edu/Research/vmd/vmd-1.9.2/ug/node89.html>`__

        Returns
        -------
        indexes : list of lists
            A list containing a list of atoms that correspond to each dihedral.
            
        Examples
        --------
        >>> dihs = []
        >>> dihs.append(Dihedral.phi(mol, 1, 2))
        >>> dihs.append(Dihedral.psi(mol, 2, 3))
        >>> indexes = Dihedral.dihedralsToIndexes(mol, dihs)
        """
        selatoms = mol.atomselect(sel)
        from moleculekit.util import ensurelist
        indexes = []
        for dih in ensurelist(dihedrals):
            idx = []
            for a in dih.atoms:
                atomsel = (mol.name == a['name']) & (mol.resid == a['resid']) & (mol.insertion == a['insertion']) & \
                          (mol.chain == a['chain']) & (mol.segid == a['segid'])
                atomsel = atomsel & selatoms
                if np.sum(atomsel) != 1:
                    raise RuntimeError(
                        'Expected one atom from atomselection {}. Got {} instead.'.format(a, np.sum(atomsel)))
                idx.append(np.where(atomsel)[0][0])
            indexes.append(idx)
        return indexes

    @staticmethod
    def _findResidue(mol, resid, insertion=None, chain=None, segid=None):
        idx = (mol.resid == resid)
        descr = 'Resid "{}"'.format(resid)
        if insertion is not None:
            idx &= (mol.insertion == insertion)
            descr += ' Insertion "{}"'.format(insertion)
        if chain is not None:
            idx &= (mol.chain == chain)
            descr += ' Chain "{}"'.format(chain)
        if segid is not None:
            idx &= (mol.segid == segid)
            descr += ' Segid "{}"'.format(segid)

        if np.sum(idx) == 0:
            raise RuntimeError('No residues found with description ({})'.format(descr))

        idx2 = np.where(idx)[0]
        if not np.array_equal(idx2, np.arange(idx2[0], idx2[-1]+1)):
            raise RuntimeError('Residue with ({}) has non-continuous indexes in PBD file ({})'.format(descr, np.where(idx)[0]))

        uqins = np.unique(mol.insertion[idx])
        if len(uqins) > 1:
            raise RuntimeError('Residue with ({}) exists with multiple insertions ({}). Define insertion to disambiguate.'.format(descr, uqins))
        uqchain = np.unique(mol.chain[idx])
        if len(uqchain) > 1:
            raise RuntimeError('Residue with ({}) exists in multiple chains ({}). Define chain to disambiguate.'.format(descr, uqchain))
        uqseg = np.unique(mol.segid[idx])
        if len(uqseg) > 1:
            raise RuntimeError('Residue with ({}) exists in multiple segments ({}). Define segid to disambiguate.'.format(descr, uqseg))
        return {'resid': resid, 'insertion': uqins[0], 'chain': uqchain[0], 'segid': uqseg[0], 'idx': idx}

    @staticmethod
    def _findAtom(mol, name, resdict):
        atomidx = resdict['idx'] & (mol.name == name)
        if np.sum(atomidx) == 0:
            raise RuntimeError('No atoms found with description ({}) and name "{}".'.format(resdict, name))
        newresdict = {'name': name, 'resid': resdict['resid'], 'insertion': resdict['insertion'], 'chain': resdict['chain'], 'segid': resdict['segid']}
        return newresdict

    @staticmethod
    def _findResname(mol, resdict):
        residx = (mol.resid == resdict['resid']) & (mol.insertion == resdict['insertion']) & \
                 (mol.chain == resdict['chain']) & (mol.segid == resdict['segid'])
        uqresname = np.unique(mol.resname[residx])

        if len(uqresname) > 1:
            raise RuntimeError('Multiple resnames ({}) found in ({})'.format(uqresname, resdict))
        return uqresname[0]

    @staticmethod
    def proteinDihedrals(mol, sel='protein', dih=('psi', 'phi')):
        """ Returns a list of tuples containing the four resid/atom pairs for each dihedral of the protein

        Parameters
        ----------
        mol : :class:`Molecule <moleculekit.molecule.Molecule>` object
            A Molecule object from which to obtain structural information
        sel : str
            Atom selection string to restrict the atoms for which to calculate dihedrals (e.g. only one of many chains).
            See more `here <http://www.ks.uiuc.edu/Research/vmd/vmd-1.9.2/ug/node89.html>`__
        dih : tuple
            A tuple of the dihedral types we want to calculate (phi, psi, omega, chi1, chi2, chi3, chi4, chi5)

        Returns
        -------
        dihedrals : list of :class:`Dihedral <moleculekit.projections.metricdihedral.Dihedral>` objects
            A list of Dihedral objects
        """
        mol = mol.copy()
        mol.filter(sel, _logger=False)
        segments = []  # Here I consider segments both chains and segments
        residues = []
        ro = io = co = so = None  # "old" values as in previous atom's
        for r, i, c, s in zip(mol.resid, mol.insertion, mol.chain, mol.segid):
            if co is not None and so is not None and ((c != co) or (s != so)):
                segments.append(residues)
                residues = []
            if (r != ro) or (i != io) or (c != co) or (s != so):
                residues.append((r, i, c, s))
                ro, io, co, so = (r, i, c, s)
        if len(residues) != 0:
            segments.append(residues)

        dihedrals = []
        for s in segments:
            residues = s
            for r in range(len(residues)):
                resid, ins, chain, segid = residues[r]
                if 'psi' in dih and r != len(residues)-1:  # No psi angle for last residue
                    resid2, ins2, _, _ = residues[r+1]
                    dihedrals.append(Dihedral.psi(mol, resid, resid2, segid, chain, ins, ins2))
                if 'phi' in dih and r != 0:  # No phi angle for first residue
                    resid1neg, ins1neg, _, _ = residues[r-1]
                    dihedrals.append(Dihedral.phi(mol, resid1neg, resid, segid, chain, ins1neg, ins))
                if 'omega' in dih and r != len(residues)-1:
                    resid2, ins2, _, _ = residues[r + 1]
                    dihedrals.append(Dihedral.omega(mol, resid, resid2, segid, chain, ins, ins2))
                if 'chi1' in dih:
                    dihedrals.append(Dihedral.chi1(mol, resid, segid, chain, ins))
                if 'chi2' in dih:
                    dihedrals.append(Dihedral.chi2(mol, resid, segid, chain, ins))
                if 'chi3' in dih:
                    dihedrals.append(Dihedral.chi3(mol, resid, segid, chain, ins))
                if 'chi4' in dih:
                    dihedrals.append(Dihedral.chi4(mol, resid, segid, chain, ins))
                if 'chi5' in dih:
                    dihedrals.append(Dihedral.chi5(mol, resid, segid, chain, ins))
        return [d for d in dihedrals if d is not None]

    # Sidechain dihedral atoms taken from
    # http://www.ccp14.ac.uk/ccp/web-mirrors/garlic/garlic/commands/dihedrals.html
    @staticmethod
    def phi(mol, res1, res2, segid=None, chain=None, insertion1=None, insertion2=None):
        """ Constructs a Dihedral object corresponding to the phi angle of res1 and res2

        Parameters
        ----------
        mol : :class:`Molecule <moleculekit.molecule.Molecule>` object
            A Molecule object from which to obtain structural information
        res1 : int
            The resid of the first residue containing the C atom
        res2 : int
            The resid of the second residue containing the N CA C atoms
        segid : str
            The segment id of the residues
        chain : str
            The chain letter of the residues
        insertion1 : str
            The insertion letter of residue 1
        insertion2 : str
            The insertion letter of residue 2

        Returns
        -------
        dihedral : :class:`Dihedral <moleculekit.projections.metricdihedral.Dihedral>` object
            A Dihedral object
        """
        res1dict = Dihedral._findResidue(mol, res1, insertion1, chain, segid)
        res2dict = Dihedral._findResidue(mol, res2, insertion2, chain, segid)
        return Dihedral(Dihedral._findAtom(mol, 'C', res1dict),
                        Dihedral._findAtom(mol, 'N', res2dict),
                        Dihedral._findAtom(mol, 'CA', res2dict),
                        Dihedral._findAtom(mol, 'C', res2dict),
                        dihedraltype='phi',
                        check_valid=False)

    @staticmethod
    def psi(mol, res1, res2, segid=None, chain=None, insertion1=None, insertion2=None):
        """ Constructs a Dihedral object corresponding to the psi angle of res1 and res2

        Parameters
        ----------
        mol : :class:`Molecule <moleculekit.molecule.Molecule>` object
            A Molecule object from which to obtain structural information
        res1 : int
            The resid of the first residue containing the N CA C atoms
        res2 : int
            The resid of the second residue containing the N atom
        segid : str
            The segment id of the residues
        chain : str
            The chain letter of the residues
        insertion1 : str
            The insertion letter of residue 1
        insertion2 : str
            The insertion letter of residue 2

        Returns
        -------
        dihedral : :class:`Dihedral <moleculekit.projections.metricdihedral.Dihedral>` object
            A Dihedral object
        """
        res1dict = Dihedral._findResidue(mol, res1, insertion1, chain, segid)
        res2dict = Dihedral._findResidue(mol, res2, insertion2, chain, segid)
        return Dihedral(Dihedral._findAtom(mol, 'N', res1dict),
                        Dihedral._findAtom(mol, 'CA', res1dict),
                        Dihedral._findAtom(mol, 'C', res1dict),
                        Dihedral._findAtom(mol, 'N', res2dict),
                        dihedraltype='psi',
                        check_valid=False)

    @staticmethod
    def omega(mol, res1, res2, segid=None, chain=None, insertion1=None, insertion2=None):
        """ Constructs a Dihedral object corresponding to the omega angle of res1 and res2

        Parameters
        ----------
        mol : :class:`Molecule <moleculekit.molecule.Molecule>` object
            A Molecule object from which to obtain structural information
        res1 : int
            The resid of the first residue containing the CA C atoms
        res2 : int
            The resid of the second residue containing the N CA atoms
        segid : str
            The segment id of the residues
        chain : str
            The chain letter of the residues
        insertion1 : str
            The insertion letter of residue 1
        insertion2 : str
            The insertion letter of residue 2

        Returns
        -------
        dihedral : :class:`Dihedral <moleculekit.projections.metricdihedral.Dihedral>` object
            A Dihedral object
        """
        res1dict = Dihedral._findResidue(mol, res1, insertion1, chain, segid)
        res2dict = Dihedral._findResidue(mol, res2, insertion2, chain, segid)
        return Dihedral(Dihedral._findAtom(mol, 'CA', res1dict),
                        Dihedral._findAtom(mol, 'C', res1dict),
                        Dihedral._findAtom(mol, 'N', res2dict),
                        Dihedral._findAtom(mol, 'CA', res2dict),
                        dihedraltype='omega',
                        check_valid=False)

    @staticmethod
    def chi1(mol, res, segid=None, chain=None, insertion=None):
        """ Constructs a Dihedral object corresponding to the chi1 angle of a residue

        Parameters
        ----------
        mol : :class:`Molecule <moleculekit.molecule.Molecule>` object
            A Molecule object from which to obtain structural information
        res : int
            The resid of the residue
        segid : str
            The segment id of the residue
        chain : str
            The chain letter of the residue
        insertion : str
            The insertion letter of the residue

        Returns
        -------
        dihedral : :class:`Dihedral <moleculekit.projections.metricdihedral.Dihedral>` object
            A Dihedral object
        """
        chi1std = ('N', 'CA', 'CB', 'CG')
        chi1 = {'ARG': chi1std, 'ASN': chi1std, 'ASP': chi1std, 'CYS': ('N', 'CA', 'CB', 'SG'), 'GLN': chi1std,
                'GLU': chi1std, 'HIS': chi1std, 'ILE': ('N', 'CA', 'CB', 'CG1'), 'LEU': chi1std, 'LYS': chi1std,
                'MET': chi1std, 'PHE': chi1std, 'PRO': chi1std, 'SER': ('N', 'CA', 'CB', 'OG'),
                'THR': ('N', 'CA', 'CB', 'OG1'), 'TRP': chi1std, 'TYR': chi1std, 'VAL': ('N', 'CA', 'CB', 'CG1')}

        resdict = Dihedral._findResidue(mol, res, insertion, chain, segid)
        resname = Dihedral._findResname(mol, resdict)
        if resname not in chi1:
            return None
        return Dihedral(Dihedral._findAtom(mol, chi1[resname][0], resdict),
                        Dihedral._findAtom(mol, chi1[resname][1], resdict),
                        Dihedral._findAtom(mol, chi1[resname][2], resdict),
                        Dihedral._findAtom(mol, chi1[resname][3], resdict),
                        dihedraltype='chi1',
                        check_valid=False)

    @staticmethod
    def chi2(mol, res, segid=None, chain=None, insertion=None):
        """ Constructs a Dihedral object corresponding to the chi2 angle of a residue

        Parameters
        ----------
        mol : :class:`Molecule <moleculekit.molecule.Molecule>` object
            A Molecule object from which to obtain structural information
        res : int
            The resid of the residue
        segid : str
            The segment id of the residue
        chain : str
            The chain letter of the residue
        insertion : str
            The insertion letter of the residue

        Returns
        -------
        dihedral : :class:`Dihedral <moleculekit.projections.metricdihedral.Dihedral>` object
            A Dihedral object
        """
        chi2std = ('CA', 'CB', 'CG', 'CD')
        chi2 = {'ARG': chi2std, 'ASN': ('CA', 'CB', 'CG', 'OD1'), 'ASP': ('CA', 'CB', 'CG', 'OD1'), 'GLN': chi2std,
                'GLU': chi2std, 'HIS': ('CA', 'CB', 'CG', 'ND1'), 'ILE': ('CA', 'CB', 'CG1', 'CD1'),
                'LEU': ('CA', 'CB', 'CG', 'CD1'), 'LYS': chi2std,
                'MET': ('CA', 'CB', 'CG', 'SD'), 'PHE': ('CA', 'CB', 'CG', 'CD1'), 'PRO': chi2std,
                'TRP': ('CA', 'CB', 'CG', 'CD1'), 'TYR': ('CA', 'CB', 'CG', 'CD1')}

        resdict = Dihedral._findResidue(mol, res, insertion, chain, segid)
        resname = Dihedral._findResname(mol, resdict)
        if resname not in chi2:
            return None
        return Dihedral(Dihedral._findAtom(mol, chi2[resname][0], resdict),
                        Dihedral._findAtom(mol, chi2[resname][1], resdict),
                        Dihedral._findAtom(mol, chi2[resname][2], resdict),
                        Dihedral._findAtom(mol, chi2[resname][3], resdict),
                        dihedraltype='chi2',
                        check_valid=False)

    @staticmethod
    def chi3(mol, res, segid=None, chain=None, insertion=None):
        """ Constructs a Dihedral object corresponding to the chi3 angle of a residue

        Parameters
        ----------
        mol : :class:`Molecule <moleculekit.molecule.Molecule>` object
            A Molecule object from which to obtain structural information
        res : int
            The resid of the residue
        segid : str
            The segment id of the residue
        chain : str
            The chain letter of the residue
        insertion : str
            The insertion letter of the residue

        Returns
        -------
        dihedral : :class:`Dihedral <moleculekit.projections.metricdihedral.Dihedral>` object
            A Dihedral object
        """
        chi3 = {'ARG': ('CB', 'CG', 'CD', 'NE'), 'GLN': ('CB', 'CG', 'CD', 'OE1'), 'GLU': ('CB', 'CG', 'CD', 'OE1'),
                'LYS': ('CB', 'CG', 'CD', 'CE'), 'MET': ('CB', 'CG', 'SD', 'CE')}

        resdict = Dihedral._findResidue(mol, res, insertion, chain, segid)
        resname = Dihedral._findResname(mol, resdict)
        if resname not in chi3:
            return None
        return Dihedral(Dihedral._findAtom(mol, chi3[resname][0], resdict),
                        Dihedral._findAtom(mol, chi3[resname][1], resdict),
                        Dihedral._findAtom(mol, chi3[resname][2], resdict),
                        Dihedral._findAtom(mol, chi3[resname][3], resdict),
                        dihedraltype='chi3',
                        check_valid=False)

    @staticmethod
    def chi4(mol, res, segid=None, chain=None, insertion=None):
        """ Constructs a Dihedral object corresponding to the chi4 angle of a residue

        Parameters
        ----------
        mol : :class:`Molecule <moleculekit.molecule.Molecule>` object
            A Molecule object from which to obtain structural information
        res : int
            The resid of the residue
        segid : str
            The segment id of the residue
        chain : str
            The chain letter of the residue
        insertion : str
            The insertion letter of the residue

        Returns
        -------
        dihedral : :class:`Dihedral <moleculekit.projections.metricdihedral.Dihedral>` object
            A Dihedral object
        """
        chi4 = {'ARG': ('CG', 'CD', 'NE', 'CZ'), 'LYS': ('CG', 'CD', 'CE', 'NZ')}

        resdict = Dihedral._findResidue(mol, res, insertion, chain, segid)
        resname = Dihedral._findResname(mol, resdict)
        if resname not in chi4:
            return None
        return Dihedral(Dihedral._findAtom(mol, chi4[resname][0], resdict),
                        Dihedral._findAtom(mol, chi4[resname][1], resdict),
                        Dihedral._findAtom(mol, chi4[resname][2], resdict),
                        Dihedral._findAtom(mol, chi4[resname][3], resdict),
                        dihedraltype='chi4',
                        check_valid=False)

    @staticmethod
    def chi5(mol, res, segid=None, chain=None, insertion=None):
        """ Constructs a Dihedral object corresponding to the chi5 angle of a residue

        Parameters
        ----------
        mol : :class:`Molecule <moleculekit.molecule.Molecule>` object
            A Molecule object from which to obtain structural information
        res : int
            The resid of the residue
        segid : str
            The segment id of the residue
        chain : str
            The chain letter of the residue
        insertion : str
            The insertion letter of the residue

        Returns
        -------
        dihedral : :class:`Dihedral <moleculekit.projections.metricdihedral.Dihedral>` object
            A Dihedral object
        """
        chi5 = {'ARG': ('CD', 'NE', 'CZ', 'NH1')}

        resdict = Dihedral._findResidue(mol, res, insertion, chain, segid)
        resname = Dihedral._findResname(mol, resdict)
        if resname not in chi5:
            return None
        return Dihedral(Dihedral._findAtom(mol, chi5[resname][0], resdict),
                        Dihedral._findAtom(mol, chi5[resname][1], resdict),
                        Dihedral._findAtom(mol, chi5[resname][2], resdict),
                        Dihedral._findAtom(mol, chi5[resname][3], resdict),
                        dihedraltype='chi5',
                        check_valid=False)


class MetricDihedral(Projection):
    """ Calculates a set of dihedral angles from trajectories

    Parameters
    ----------
    dih : list of :class:`Dihedral <moleculekit.projections.metricdihedral.Dihedral>` object
        You can provide your own list of Dihedral objects. See example.
    sincos : bool, optional
        Set to True to return the dihedral angles as their sine and cosine components. Makes them periodic.
    protsel : str, optional
        Atom selection string for the protein segment for which to calculate dihedral angles. Resids should be unique
        within that segment. See more `here <http://www.ks.uiuc.edu/Research/vmd/vmd-1.9.2/ug/node89.html>`__

    Examples
    --------
    >>> mol = Molecule('3PTB')
    >>> mol.filter('not insertion A')
    >>> met = MetricDihedral()
    >>> met.project(mol)
    >>> # More complicated example
    >>> dih = []
    >>> dih.append(Dihedral.chi1(mol, 45))
    >>> dih.append(Dihedral.psi(mol, 29, 30))
    >>> met = MetricDihedral(dih, protsel='protein and segid 0')
    >>> met.project(mol)
    >>> met.getMapping(mol)
    """
    def __init__(self, dih=None, sincos=True, protsel='protein'):
        super().__init__()

        if dih is not None and not isinstance(dih[0], Dihedral):
            raise RuntimeError('Manually passing dihedrals to MetricDihedral requires use of the Dihedral class. Check the example in the documentation')
        
        self._protsel = protsel
        self._sincos = sincos
        self._dihedrals = dih

    def project(self, mol):
        """ Project molecule.

        Parameters
        ----------
        mol : :class:`Molecule <moleculekit.molecule.Molecule>`
            A :class:`Molecule <moleculekit.molecule.Molecule>` object to project.

        Returns
        -------
        data : np.ndarray
            An array containing the projected data.
        """
        dihedrals = self._getMolProp(mol, 'dihedrals')
        return self._calcDihedralAngles(mol, dihedrals, sincos=self._sincos)

    def getMapping(self, mol):
        """ Returns the description of each projected dimension.

        Parameters
        ----------
        mol : :class:`Molecule <moleculekit.molecule.Molecule>` object
            A Molecule object which will be used to calculate the descriptions of the projected dimensions.

        Returns
        -------
        map : :class:`DataFrame <pandas.core.frame.DataFrame>` object
            A DataFrame containing the descriptions of each dimension
        """
        dihedrals = self._getMolProp(mol, 'dihedrals')

        from pandas import DataFrame
        types = []
        indexes = []
        description = []
        for i, dih in enumerate(dihedrals):
            types += ['dihedral']
            indexes += [dih]
            mapstr = ''
            for d in dih:
                mapstr += '({} {} {} {} {}) '.format(mol.resname[d], mol.resid[d], mol.name[d], mol.segid[d], mol.chain[d])
            if self._sincos:
                description += ['Sine of angle of ' + mapstr]
                description += ['Cosine of angle of ' + mapstr]
                types += ['dihedral']
                indexes += [dih]
            else:
                description += ['Angle of ' + mapstr]
        return DataFrame({'type': types, 'atomIndexes': indexes, 'description': description})

    def _calculateMolProp(self, mol, props='all'):
        from moleculekit.util import ensurelist
        res = {}
        protsel = mol.atomselect(self._protsel)

        if self._dihedrals is None:  # Default phi psi dihedrals
            dihedrals = Dihedral.proteinDihedrals(mol, protsel)
        else:
            dihedrals = ensurelist(self._dihedrals)

        res['dihedrals'] = Dihedral.dihedralsToIndexes(mol, dihedrals, protsel)
        return res

    def _calcDihedralAngles(self, mol, dihedrals, sincos=True):
        from moleculekit.dihedral import dihedralAngle
        metric = np.zeros((np.size(mol.coords, 2), len(dihedrals)))

        for i, dih in enumerate(dihedrals):
            metric[:, i] = np.rad2deg(dihedralAngle(mol.coords[dih, :, :]))

        if sincos:
            sc_metric = np.zeros((np.size(metric, 0), np.size(metric, 1) * 2))
            sc_metric[:, 0::2] = np.sin(metric * np.pi / 180.)
            sc_metric[:, 1::2] = np.cos(metric * np.pi / 180.)
            metric = sc_metric
        return metric.astype(np.float32)


import unittest
class _TestMetricDihedral(unittest.TestCase):
    def test_dihedral_traj(self):
        from moleculekit.molecule import Molecule
        from moleculekit.home import home
        from os import path

        mol = Molecule(path.join(home(dataDir='test-projections'), 'trajectory', 'filtered.pdb'))
        mol.read(path.join(home(dataDir='test-projections'), 'trajectory', 'traj.xtc'))

        metr = MetricDihedral(protsel='protein')
        data = metr.project(mol)
        dataref = np.load(path.join(home(dataDir='test-projections'), 'metricdihedral', 'ref.npy'))
        assert np.allclose(data, dataref, atol=1e-03), 'Diherdals calculation gave different results from reference'

    def test_dihedral_5mat(self):
        from moleculekit.molecule import Molecule
        from moleculekit.home import home
        from moleculekit.tools.autosegment import autoSegment
        from os import path

        mol = Molecule('5MAT')
        mol.filter('not insertion A and not altloc A B', _logger=False)
        mol = autoSegment(mol, _logger=False)
        data = MetricDihedral().project(mol)
        dataref = np.load(path.join(home(dataDir='test-projections'), 'metricdihedral', '5mat.npy'))
        assert np.allclose(data, dataref, atol=1e-03), 'Diherdals calculation gave different results from reference'


if __name__ == '__main__':
    unittest.main(verbosity=2)



# (c) 2015-2018 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#
import os

_config = {'viewer': 'VMD',
           'configfile': os.getenv('HTMD_CONFIG') if os.getenv('HTMD_CONFIG') else None,
           'ncpus': 1
           }


def config(viewer=_config['viewer'],
           configfile=_config['configfile'],
           ncpus=_config['ncpus']):
    """
    Function to change HTMD configuration variables.

    Parameters
    ----------
    viewer : str
        Defines the backend viewer for molecular visualization
    ncpus : int
        Defines the number of cpus available for several HTMD operations
    configfile : str
        Defines the HTMD configuration file that is called at the beginning of importing
    lsf : str
        Defines a YAML file that can contain default profile configurations for an LsfQueue
    slurm : str
        Defines a YAML file that can contain default profile configurations for an SlurmQueue
    """
    _config['viewer'] = viewer
    _config['configfile'] = configfile
    _config['ncpus'] = ncpus


from collections import namedtuple

# from periodictable import elements  # the python package called periodictable, not this file
# for el in elements:
#     if el.name == 'neutron':
#         continue
#     # if mendeleev.__dict__[el.symbol].vdw_radius_bondi is not None:
#     #     vdw_radius = mendeleev.__dict__[el.symbol].vdw_radius_bondi
#     # else:
#     vdw_radius = mendeleev.__dict__[el.symbol].vdw_radius
#     if vdw_radius is not None:
#         vdw_radius /= 100  # Convert picometer to A
#     print("'{}': _Element(symbol='{}', name='{}', number={}, mass={}, vdw_radius={}),".format(el, el, el.name, el.number, el.mass, vdw_radius))


_Element = namedtuple('Element', ['symbol', 'name', 'number', 'mass', 'vdw_radius'])
periodictable = {
    'H': _Element(symbol='H', name='hydrogen', number=1, mass=1.00794, vdw_radius=1.1),
    'He': _Element(symbol='He', name='helium', number=2, mass=4.002602, vdw_radius=1.4),
    'Li': _Element(symbol='Li', name='lithium', number=3, mass=6.941, vdw_radius=1.82),
    'Be': _Element(symbol='Be', name='beryllium', number=4, mass=9.012182, vdw_radius=1.53),
    'B': _Element(symbol='B', name='boron', number=5, mass=10.811, vdw_radius=1.92),
    'C': _Element(symbol='C', name='carbon', number=6, mass=12.0107, vdw_radius=1.7),
    'N': _Element(symbol='N', name='nitrogen', number=7, mass=14.0067, vdw_radius=1.55),
    'O': _Element(symbol='O', name='oxygen', number=8, mass=15.9994, vdw_radius=1.52),
    'F': _Element(symbol='F', name='fluorine', number=9, mass=18.9984032, vdw_radius=1.47),
    'Ne': _Element(symbol='Ne', name='neon', number=10, mass=20.1797, vdw_radius=1.54),
    'Na': _Element(symbol='Na', name='sodium', number=11, mass=22.98977, vdw_radius=2.27),
    'Mg': _Element(symbol='Mg', name='magnesium', number=12, mass=24.305, vdw_radius=1.73),
    'Al': _Element(symbol='Al', name='aluminum', number=13, mass=26.981538, vdw_radius=1.84),
    'Si': _Element(symbol='Si', name='silicon', number=14, mass=28.0855, vdw_radius=2.1),
    'P': _Element(symbol='P', name='phosphorus', number=15, mass=30.973761, vdw_radius=1.8),
    'S': _Element(symbol='S', name='sulfur', number=16, mass=32.065, vdw_radius=1.8),
    'Cl': _Element(symbol='Cl', name='chlorine', number=17, mass=35.453, vdw_radius=1.75),
    'Ar': _Element(symbol='Ar', name='argon', number=18, mass=39.948, vdw_radius=1.88),
    'K': _Element(symbol='K', name='potassium', number=19, mass=39.0983, vdw_radius=2.75),
    'Ca': _Element(symbol='Ca', name='calcium', number=20, mass=40.078, vdw_radius=2.31),
    'Sc': _Element(symbol='Sc', name='scandium', number=21, mass=44.95591, vdw_radius=2.15),
    'Ti': _Element(symbol='Ti', name='titanium', number=22, mass=47.867, vdw_radius=2.11),
    'V': _Element(symbol='V', name='vanadium', number=23, mass=50.9415, vdw_radius=2.07),
    'Cr': _Element(symbol='Cr', name='chromium', number=24, mass=51.9961, vdw_radius=2.06),
    'Mn': _Element(symbol='Mn', name='manganese', number=25, mass=54.938049, vdw_radius=2.05),
    'Fe': _Element(symbol='Fe', name='iron', number=26, mass=55.845, vdw_radius=2.04),
    'Co': _Element(symbol='Co', name='cobalt', number=27, mass=58.9332, vdw_radius=2.0),
    'Ni': _Element(symbol='Ni', name='nickel', number=28, mass=58.6934, vdw_radius=1.97),
    'Cu': _Element(symbol='Cu', name='copper', number=29, mass=63.546, vdw_radius=1.96),
    'Zn': _Element(symbol='Zn', name='zinc', number=30, mass=65.409, vdw_radius=2.01),
    'Ga': _Element(symbol='Ga', name='gallium', number=31, mass=69.723, vdw_radius=1.87),
    'Ge': _Element(symbol='Ge', name='germanium', number=32, mass=72.64, vdw_radius=2.11),
    'As': _Element(symbol='As', name='arsenic', number=33, mass=74.9216, vdw_radius=1.85),
    'Se': _Element(symbol='Se', name='selenium', number=34, mass=78.96, vdw_radius=1.9),
    'Br': _Element(symbol='Br', name='bromine', number=35, mass=79.904, vdw_radius=1.85),
    'Kr': _Element(symbol='Kr', name='krypton', number=36, mass=83.798, vdw_radius=2.02),
    'Rb': _Element(symbol='Rb', name='rubidium', number=37, mass=85.4678, vdw_radius=3.03),
    'Sr': _Element(symbol='Sr', name='strontium', number=38, mass=87.62, vdw_radius=2.49),
    'Y': _Element(symbol='Y', name='yttrium', number=39, mass=88.90585, vdw_radius=2.32),
    'Zr': _Element(symbol='Zr', name='zirconium', number=40, mass=91.224, vdw_radius=2.23),
    'Nb': _Element(symbol='Nb', name='niobium', number=41, mass=92.90638, vdw_radius=2.18),
    'Mo': _Element(symbol='Mo', name='molybdenum', number=42, mass=95.94, vdw_radius=2.17),
    'Tc': _Element(symbol='Tc', name='technetium', number=43, mass=98, vdw_radius=2.16),
    'Ru': _Element(symbol='Ru', name='ruthenium', number=44, mass=101.07, vdw_radius=2.13),
    'Rh': _Element(symbol='Rh', name='rhodium', number=45, mass=102.9055, vdw_radius=2.1),
    'Pd': _Element(symbol='Pd', name='palladium', number=46, mass=106.42, vdw_radius=2.1),
    'Ag': _Element(symbol='Ag', name='silver', number=47, mass=107.8682, vdw_radius=2.11),
    'Cd': _Element(symbol='Cd', name='cadmium', number=48, mass=112.411, vdw_radius=2.18),
    'In': _Element(symbol='In', name='indium', number=49, mass=114.818, vdw_radius=1.93),
    'Sn': _Element(symbol='Sn', name='tin', number=50, mass=118.71, vdw_radius=2.17),
    'Sb': _Element(symbol='Sb', name='antimony', number=51, mass=121.76, vdw_radius=2.06),
    'Te': _Element(symbol='Te', name='tellurium', number=52, mass=127.6, vdw_radius=2.06),
    'I': _Element(symbol='I', name='iodine', number=53, mass=126.90447, vdw_radius=1.98),
    'Xe': _Element(symbol='Xe', name='xenon', number=54, mass=131.293, vdw_radius=2.16),
    'Cs': _Element(symbol='Cs', name='cesium', number=55, mass=132.90545, vdw_radius=3.43),
    'Ba': _Element(symbol='Ba', name='barium', number=56, mass=137.327, vdw_radius=2.68),
    'La': _Element(symbol='La', name='lanthanum', number=57, mass=138.9055, vdw_radius=2.43),
    'Ce': _Element(symbol='Ce', name='cerium', number=58, mass=140.116, vdw_radius=2.42),
    'Pr': _Element(symbol='Pr', name='praseodymium', number=59, mass=140.90765, vdw_radius=2.4),
    'Nd': _Element(symbol='Nd', name='neodymium', number=60, mass=144.24, vdw_radius=2.39),
    'Pm': _Element(symbol='Pm', name='promethium', number=61, mass=145, vdw_radius=2.38),
    'Sm': _Element(symbol='Sm', name='samarium', number=62, mass=150.36, vdw_radius=2.36),
    'Eu': _Element(symbol='Eu', name='europium', number=63, mass=151.964, vdw_radius=2.35),
    'Gd': _Element(symbol='Gd', name='gadolinium', number=64, mass=157.25, vdw_radius=2.34),
    'Tb': _Element(symbol='Tb', name='terbium', number=65, mass=158.92534, vdw_radius=2.33),
    'Dy': _Element(symbol='Dy', name='dysprosium', number=66, mass=162.5, vdw_radius=2.31),
    'Ho': _Element(symbol='Ho', name='holmium', number=67, mass=164.93032, vdw_radius=2.3),
    'Er': _Element(symbol='Er', name='erbium', number=68, mass=167.259, vdw_radius=2.29),
    'Tm': _Element(symbol='Tm', name='thulium', number=69, mass=168.93421, vdw_radius=2.27),
    'Yb': _Element(symbol='Yb', name='ytterbium', number=70, mass=173.04, vdw_radius=2.26),
    'Lu': _Element(symbol='Lu', name='lutetium', number=71, mass=174.967, vdw_radius=2.24),
    'Hf': _Element(symbol='Hf', name='hafnium', number=72, mass=178.49, vdw_radius=2.23),
    'Ta': _Element(symbol='Ta', name='tantalum', number=73, mass=180.9479, vdw_radius=2.22),
    'W': _Element(symbol='W', name='tungsten', number=74, mass=183.84, vdw_radius=2.18),
    'Re': _Element(symbol='Re', name='rhenium', number=75, mass=186.207, vdw_radius=2.16),
    'Os': _Element(symbol='Os', name='osmium', number=76, mass=190.23, vdw_radius=2.16),
    'Ir': _Element(symbol='Ir', name='iridium', number=77, mass=192.217, vdw_radius=2.13),
    'Pt': _Element(symbol='Pt', name='platinum', number=78, mass=195.078, vdw_radius=2.13),
    'Au': _Element(symbol='Au', name='gold', number=79, mass=196.96655, vdw_radius=2.14),
    'Hg': _Element(symbol='Hg', name='mercury', number=80, mass=200.59, vdw_radius=2.23),
    'Tl': _Element(symbol='Tl', name='thallium', number=81, mass=204.3833, vdw_radius=1.96),
    'Pb': _Element(symbol='Pb', name='lead', number=82, mass=207.2, vdw_radius=2.02),
    'Bi': _Element(symbol='Bi', name='bismuth', number=83, mass=208.98038, vdw_radius=2.07),
    'Po': _Element(symbol='Po', name='polonium', number=84, mass=209, vdw_radius=1.97),
    'At': _Element(symbol='At', name='astatine', number=85, mass=210, vdw_radius=2.02),
    'Rn': _Element(symbol='Rn', name='radon', number=86, mass=222, vdw_radius=2.2),
    'Fr': _Element(symbol='Fr', name='francium', number=87, mass=223, vdw_radius=3.48),
    'Ra': _Element(symbol='Ra', name='radium', number=88, mass=226, vdw_radius=2.83),
    'Ac': _Element(symbol='Ac', name='actinium', number=89, mass=227, vdw_radius=2.47),
    'Th': _Element(symbol='Th', name='thorium', number=90, mass=232.0381, vdw_radius=2.45),
    'Pa': _Element(symbol='Pa', name='protactinium', number=91, mass=231.03588, vdw_radius=2.43),
    'U': _Element(symbol='U', name='uranium', number=92, mass=238.02891, vdw_radius=2.41),
    'Np': _Element(symbol='Np', name='neptunium', number=93, mass=237, vdw_radius=2.39),
    'Pu': _Element(symbol='Pu', name='plutonium', number=94, mass=244, vdw_radius=2.43),
    'Am': _Element(symbol='Am', name='americium', number=95, mass=243, vdw_radius=2.44),
    'Cm': _Element(symbol='Cm', name='curium', number=96, mass=247, vdw_radius=2.45),
    'Bk': _Element(symbol='Bk', name='berkelium', number=97, mass=247, vdw_radius=2.44),
    'Cf': _Element(symbol='Cf', name='californium', number=98, mass=251, vdw_radius=2.45),
    'Es': _Element(symbol='Es', name='einsteinium', number=99, mass=252, vdw_radius=2.45),
    'Fm': _Element(symbol='Fm', name='fermium', number=100, mass=257, vdw_radius=2.45),
    'Md': _Element(symbol='Md', name='mendelevium', number=101, mass=258, vdw_radius=2.46),
    'No': _Element(symbol='No', name='nobelium', number=102, mass=259, vdw_radius=2.46),
    'Lr': _Element(symbol='Lr', name='lawrencium', number=103, mass=262, vdw_radius=2.46),
    'Rf': _Element(symbol='Rf', name='rutherfordium', number=104, mass=261, vdw_radius=None),
    'Db': _Element(symbol='Db', name='dubnium', number=105, mass=262, vdw_radius=None),
    'Sg': _Element(symbol='Sg', name='seaborgium', number=106, mass=266, vdw_radius=None),
    'Bh': _Element(symbol='Bh', name='bohrium', number=107, mass=264, vdw_radius=None),
    'Hs': _Element(symbol='Hs', name='hassium', number=108, mass=277, vdw_radius=None),
    'Mt': _Element(symbol='Mt', name='meitnerium', number=109, mass=268, vdw_radius=None),
    'Ds': _Element(symbol='Ds', name='darmstadtium', number=110, mass=281, vdw_radius=None),
    'Rg': _Element(symbol='Rg', name='roentgenium', number=111, mass=272, vdw_radius=None),
    'Cn': _Element(symbol='Cn', name='copernicium', number=112, mass=285, vdw_radius=None),
    'Nh': _Element(symbol='Nh', name='nihonium', number=113, mass=286, vdw_radius=None),
    'Fl': _Element(symbol='Fl', name='flerovium', number=114, mass=289, vdw_radius=None),
    'Mc': _Element(symbol='Mc', name='moscovium', number=115, mass=289, vdw_radius=None),
    'Lv': _Element(symbol='Lv', name='livermorium', number=116, mass=293, vdw_radius=None),
    'Ts': _Element(symbol='Ts', name='tennessine', number=117, mass=294, vdw_radius=None),
    'Og': _Element(symbol='Og', name='oganesson', number=118, mass=294, vdw_radius=None),
}



import string
import numpy as np
import logging


logger = logging.getLogger(__name__)

def autoSegment(mol, sel='all', basename='P', spatial=True, spatialgap=4.0, field="segid", _logger=True):
    """ Detects resid gaps in a selection and assigns incrementing segid to each fragment

    !!!WARNING!!! If you want to use atom selections like 'protein' or 'fragment',
    use this function on a Molecule containing only protein atoms, otherwise the protein selection can fail.

    Parameters
    ----------
    mol : :class:`Molecule <moleculekit.molecule.Molecule>` object
        The Molecule object
    sel : str
        Atom selection string on which to check for gaps.
        See more `here <http://www.ks.uiuc.edu/Research/vmd/vmd-1.9.2/ug/node89.html>`__
    basename : str
        The basename for segment ids. For example if given 'P' it will name the segments 'P1', 'P2', ...
    spatial : bool
        Only considers a discontinuity in resid as a gap of the CA atoms have distance more than `spatialgap` Angstrom
    spatialgap : float
        The size of a spatial gap which validates a discontinuity (A)
    field : str
        Field to fix. Can be "segid" (default), "chain", or "both"

    Returns
    -------
    newmol : :class:`Molecule <moleculekit.molecule.Molecule>` object
        A new Molecule object with modified segids

    Example
    -------
    >>> newmol = autoSegment(mol,'chain B','P')
    """
    from moleculekit.util import sequenceID
    mol = mol.copy()

    idx = mol.atomselect(sel, indexes=True)
    rid = mol.get('resid', sel)
    residiff = np.diff(rid)
    gappos = np.where((residiff != 1) & (residiff != 0))[0]  # Points to the index before the gap!

    # Letters to be used for chains, if free: 0123456789abcd...ABCD..., minus chain symbols already used
    used_chains = set(mol.chain)
    chain_alphabet = list(string.digits + string.ascii_letters)
    available_chains = [x for x in chain_alphabet if x not in used_chains]

    idxstartseg = [idx[0]] + idx[gappos + 1].tolist()
    idxendseg = idx[gappos].tolist() + [idx[-1]]

    mol.set('segid', basename, sel)

    if len(gappos) == 0:
        mol.set('segid', basename+'0', sel)
        return mol

    if spatial:
        residbackup = mol.get('resid')
        mol.set('resid', sequenceID(mol.resid))  # Assigning unique resids to be able to do the distance selection

        todelete = []
        i = 0
        for s, e in zip(idxstartseg[1:], idxendseg[:-1]):
            coords = mol.get('coords', sel='resid "{}" "{}" and name CA'.format(mol.resid[e], mol.resid[s]))
            if np.shape(coords) == (2, 3):
                dist = np.sqrt(np.sum((coords[0, :] - coords[1, :]) ** 2))
                if dist < spatialgap:
                    todelete.append(i)
            i += 1
        todelete = np.array(todelete, dtype=int)
        # Join the non-real gaps into segments
        idxstartseg = np.delete(idxstartseg, todelete+1)
        idxendseg = np.delete(idxendseg, todelete)

        mol.set('resid', residbackup)  # Restoring the original resids

    i = 0
    for s, e in zip(idxstartseg, idxendseg):
        # Fixup segid
        if field in ['segid', 'both']:
            newsegid = basename + str(i)
            if np.any(mol.segid == newsegid):
                raise RuntimeError('Segid {} already exists in the molecule. Please choose different prefix.'.format(newsegid))
            if _logger: logger.info('Created segment {} between resid {} and {}.'.format(newsegid, mol.resid[s], mol.resid[e]))
            mol.segid[s:e+1] = newsegid
        # Fixup chain
        if field in ['chain', 'both']:
            newchainid = available_chains[i]
            if _logger: logger.info('Set chain {} between resid {} and {}.'.format(newchainid, mol.resid[s], mol.resid[e]))
            mol.chain[s:e+1] = newchainid

        i += 1

    return mol


def autoSegment2(mol, sel='(protein or resname ACE NME)', basename='P', fields=('segid',), residgaps=False, residgaptol=1, chaingaps=True, _logger=True):
    """ Detects bonded segments in a selection and assigns incrementing segid to each segment

    Parameters
    ----------
    mol : :class:`Molecule <moleculekit.molecule.Molecule>` object
        The Molecule object
    sel : str
        Atom selection string on which to check for gaps.
        See more `here <http://www.ks.uiuc.edu/Research/vmd/vmd-1.9.2/ug/node89.html>`__
    basename : str
        The basename for segment ids. For example if given 'P' it will name the segments 'P1', 'P2', ...
    fields : tuple of strings
        Field to fix. Can be "segid" (default) or any other Molecule field or combinations thereof.
    residgaps : bool
        Set to True to consider gaps in resids as structural gaps. Set to False to ignore resids
    residgaptol : int
        Above what resid difference is considered a gap. I.e. with residgaptol 1, 235-233 = 2 > 1 hence is a gap. We set
        default to 2 because in many PDBs single residues are missing in the proteins without any gaps.
    chaingaps : bool
        Set to True to consider changes in chains as structural gaps. Set to False to ignore chains

    Returns
    -------
    newmol : :class:`Molecule <moleculekit.molecule.Molecule>` object
        A new Molecule object with modified segids

    Example
    -------
    >>> newmol = autoSegment2(mol)
    """
    from scipy.sparse import csr_matrix
    from scipy.sparse.csgraph import connected_components

    if isinstance(fields, str):
        fields = (fields,)

    sel += ' and backbone or (resname NME ACE and name N C O CH3)'  # Looking for bonds only over the backbone of the protein
    idx = mol.atomselect(sel, indexes=True)  # Keep the original atom indexes to map from submol to mol
    submol = mol.copy()  # We filter out everything not on the backbone to calculate only those bonds
    submol.filter(sel, _logger=False)
    bonds = submol._getBonds()  # Calculate both file and guessed bonds

    if residgaps:
        # Remove bonds between residues without continuous resids
        bondresiddiff = np.abs(submol.resid[bonds[:, 0]] - submol.resid[bonds[:, 1]])
        bonds = bonds[bondresiddiff <= residgaptol, :]
    else:
        # Warning about bonds bonding non-continuous resids
        bondresiddiff = np.abs(submol.resid[bonds[:, 0]] - submol.resid[bonds[:, 1]])
        if _logger and np.any(bondresiddiff > 1):
            for i in np.where(bondresiddiff > residgaptol)[0]:
                logger.warning('Bonds found between resid gaps: resid {} and {}'.format(submol.resid[bonds[i, 0]],
                                                                                        submol.resid[bonds[i, 1]]))
    if chaingaps:
        # Remove bonds between residues without same chain
        bondsamechain = submol.chain[bonds[:, 0]] == submol.chain[bonds[:, 1]]
        bonds = bonds[bondsamechain, :]
    else:
        # Warning about bonds bonding different chains
        bondsamechain = submol.chain[bonds[:, 0]] == submol.chain[bonds[:, 1]]
        if _logger and np.any(bondsamechain == False):
            for i in np.where(bondsamechain == False)[0]:
                logger.warning('Bonds found between chain gaps: resid {}/{} and {}/{}'.format(submol.resid[bonds[i, 0]],
                                                                                              submol.chain[bonds[i, 0]],
                                                                                              submol.resid[bonds[i, 1]],
                                                                                              submol.chain[bonds[i, 1]]
                                                                                              ))

    # Calculate connected components using the bonds
    sparsemat = csr_matrix((np.ones(bonds.shape[0] * 2),  # Values
                            (np.hstack((bonds[:, 0], bonds[:, 1])),  # Rows
                             np.hstack((bonds[:, 1], bonds[:, 0])))), shape=[submol.numAtoms, submol.numAtoms])  # Columns
    numcomp, compidx = connected_components(sparsemat, directed=False)

    # Letters to be used for chains, if free: 0123456789abcd...ABCD..., minus chain symbols already used
    used_chains = set(mol.chain)
    chain_alphabet = list(string.digits + string.ascii_letters)
    available_chains = [x for x in chain_alphabet if x not in used_chains]

    mol = mol.copy()
    prevsegres = None
    for i in range(numcomp):  # For each connected component / segment
        segid = basename + str(i)
        backboneSegIdx = idx[compidx == i]  # The backbone atoms of the segment
        segres = mol.atomselect('same residue as index {}'.format(' '.join(map(str, backboneSegIdx)))) # Get whole residues

        # Warning about separating segments with continuous resids
        if _logger and i > 0 and (np.min(mol.resid[segres]) - np.max(mol.resid[prevsegres])) == 1:
            logger.warning('Separated segments {} and {}, despite continuous resids, due to lack of bonding.'.format(
                            basename + str(i-1), segid))

        # Add the new segment ID to all fields the user specified
        for f in fields:
            if f != 'chain':
                if np.any(mol.__dict__[f] == segid):
                    raise RuntimeError('Segid {} already exists in the molecule. Please choose different prefix.'.format(segid))
                mol.__dict__[f][segres] = segid  # Assign the segid to the correct atoms
            else:
                mol.__dict__[f][segres] = available_chains[i % len(available_chains)]
        if _logger:
            logger.info('Created segment {} between resid {} and {}.'.format(segid, np.min(mol.resid[segres]),
                                                                            np.max(mol.resid[segres])))
        prevsegres = segres  # Store old segment atom indexes for the warning about continuous resids

    return mol

import unittest
class _TestPreparation(unittest.TestCase):
    def test_autoSegment(self):
        from moleculekit.home import home
        from moleculekit.molecule import Molecule
        from os import path

        p = Molecule(path.join(home(dataDir='test-autosegment'), '4dkl.pdb'))
        p.filter('(chain B and protein) or water')
        p = autoSegment(p, 'protein', 'P')
        m = Molecule(path.join(home(dataDir='test-autosegment'), 'membrane.pdb'))
        print(np.unique(m.get('segid')))

        mol = Molecule(path.join(home(dataDir='test-autosegment'), '1ITG_clean.pdb'))
        ref = Molecule(path.join(home(dataDir='test-autosegment'), '1ITG.pdb'))
        mol = autoSegment(mol, sel='protein')
        assert np.all(mol.segid == ref.segid)

        mol = Molecule(path.join(home(dataDir='test-autosegment'), '3PTB_clean.pdb'))
        ref = Molecule(path.join(home(dataDir='test-autosegment'), '3PTB.pdb'))
        mol = autoSegment(mol, sel='protein')
        assert np.all(mol.segid == ref.segid)

if __name__ == "__main__":
    unittest.main(verbosity=2)


# (c) 2015-2018 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#
import tempfile
import os
import re
import numpy as np

try:
    import pandas
except ImportError as e:
    print('Pandas is not installed. Please install it with `conda install pandas`')()


def getProfile(sequence, hhblits, hhblitsdb, ncpu=6, niter=4):
    """ Calculates the sequence profile of a protein sequence using HHBlits

    File description and unit conversions taken from section 6 https://hpc.nih.gov/apps/hhsuite-userguide.pdf

    Parameters
    ----------
    sequence : str
        A string encoding the one letter sequence of the protein
    hhblits : str
        The path to the hhblits executable
    hhblitsdb : str
        The path to the hhblits database that we want to search against
    ncpu : int
        Number of CPUs to use for the search
    niter : int
        The number of hhblits iterations. The higher the value the more remote homologues it will find
    
    Returns
    -------
    df : pandas.DataFrame
        A pandas dataframe containing all the information read from the file
    pssm : np.ndarray
        A Nx20 numpy array where N the number of residues of the protein. Contains the transition probabilities to all 20 residues.

    Examples
    --------
    >>> hhb = '~/hhsuite-2.0.16-linux-x86_64/bin/hhblits'
    >>> hhbdb = '~/hhsuite-2.0.16-linux-x86_64/databases/uniprot20_2016_02/uniprot20_2016_02'
    >>> seq = 'MKVIFLKDVKGMGKKGEIKNVADGYANNFLFKQGLAIEA'
    >>> df, prof = getProfile(seq, hhb, hhbdb)
    """
    regex = re.compile('^\w\s\d+')
    with tempfile.TemporaryDirectory() as tmpdir:
        with open(tmpdir + '/input.fasta', 'w') as fp:
            fp.write('>\n{}\n'.format(sequence))

        os.system('{} -i {}/input.fasta -d {} -cpu {} -M first -n {} -ohhm {}/output.hhm -v 0'.format(hhblits, tmpdir, hhblitsdb, ncpu, niter, tmpdir))
        
        data = []
        seq = []
        with open(tmpdir + '/output.hhm', 'r') as fp:
            starting = 0
            lines = fp.readlines()
            for i in range(len(lines)):
                if lines[i].startswith('NULL'):
                    pieces = lines[i].split()
                    seq.append([pieces[0]])
                    data.append([2 ** (-int(x) / 1000) if x != '*' else 0 for x in pieces[1:21]] + [0] * 10)
                if lines[i].startswith('HMM    A	C	D'):
                    col_desc = lines[i].split()[1:] + lines[i + 1].split()
                    starting = 1
                if starting > 0:
                    starting += 1
                if starting >= 4 and regex.match(lines[i]):
                    pieces = lines[i].split()
                    seq.append([pieces[0]])
                    d = [2 ** (-int(x) / 1000) if x != '*' else 0 for x in pieces[2:22]]
                    pieces = lines[i + 1].split()
                    d += [2 ** (-int(x) / 1000) if x != '*' else 0 for x in pieces[:7]]
                    d += [0.001 * int(x) for x in pieces[7:10]]
                    data.append(d)

        df = pandas.DataFrame(np.hstack((np.vstack(seq), np.vstack(data))), columns=['seq'] + col_desc)
        return df, df.values[1:, 1:21].astype(float)


if __name__ == '__main__':
    pass



# (c) 2015-2018 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#
import os
import ctypes as ct

import numpy
import numpy as np

from moleculekit.home import home


def wrap(coordinates, bonds, box, centersel=None):
    """Wrap the coordinates back into the unit cell. Molecules will remain continuous, so may escape the bounds of the prinary unit cell.

    Parameters
    ----------

    coordinates :
    bonds       :
    box         :

    Return
    ------
 
    coordinates  
    """

    import platform
    libdir = home(libDir=True)

    if coordinates.dtype != np.float32:
        raise ValueError("Coordinates is not float32")

    if coordinates.ndim == 2:
        c = coordinates.shape
        coordinates = coordinates.reshape((c[0], c[1], 1))

    if coordinates.shape[1] != 3:
        #    print(coordinates.shape)
        raise NameError("Coordinates needs to be natoms x 3 x nframes")

    z = coordinates.shape[2]
    if coordinates.strides[0] != 12 * z or coordinates.strides[1] != 4 * z:
        # It's a view -- need to make a copy to ensure contiguity of memory
        coordinates = numpy.array(coordinates, dtype=numpy.float32)
    if coordinates.strides[0] != 12 * z or coordinates.strides[1] != 4 * z:
        raise ValueError("Coordinates is a view with unsupported strides")

    natoms = coordinates.shape[0]
    nframes = coordinates.shape[2]
    # print(box.shape)
    if numpy.size(bonds, 1) != 2:
        raise NameError("'bonds' not nbonds x 2 in length")
    if numpy.size(box, 0) != 3:
        raise NameError("'box' not nframes x 3 in length")
    if numpy.size(box, 1) != nframes:
        raise NameError("'box' not nframes x 3 in length")

    if platform.system() == "Windows":
        ct.cdll.LoadLibrary(os.path.join(libdir, "libgcc_s_seh-1.dll"))
        if os.path.exists(os.path.join(libdir, "psprolib.dll")):
            ct.cdll.LoadLibrary(os.path.join(libdir, "psprolib.dll"))

    lib = ct.cdll.LoadLibrary(os.path.join(libdir, "libvmdparser.so"))

    nbonds = bonds.shape[0]
    ll = 3 * nbonds
    c_bonds = (ct.c_int * ll)()

    for z in range(0, nbonds):
        for y in [0, 1]:
            c_bonds[z * 2 + y] = bonds[z, y]

    ll = 3 * nframes
    c_box = (ct.c_double * ll)()

    z = 0
    for i in range(0, nframes):
        for j in range(0, 3):
            c_box[z] = box[j][i]
            z = z + 1

    c_nbonds = ct.c_int(nbonds)
    c_natoms = ct.c_int(natoms)
    c_nframes = ct.c_int(nframes)
    lenv = natoms * 3 * nframes

    if centersel is None:
        centersel = numpy.array([-1], dtype=numpy.int32)
    centersel = numpy.append(centersel, numpy.array([-1], dtype=numpy.int32))
    c_centersel = centersel.ctypes.data_as(ct.POINTER(ct.c_int))
    c_coords = coordinates.ctypes.data_as(ct.POINTER(ct.c_float))
    lib.wrap(
        c_bonds,
        c_coords,
        c_box, c_nbonds, c_natoms, c_nframes, c_centersel
    )

    return coordinates


from unittest import TestCase
class _TestWrap(TestCase):
    def test_wrap(self):
        from moleculekit.molecule import Molecule
        from moleculekit.home import home

        datadir = os.path.join(home(dataDir='molecule-readers'), 'multi-traj')
        mol = Molecule(os.path.join(datadir, 'structure.pdb'))
        mol.read(os.path.join(datadir, 'data', 'e1s1_1', 'output.xtc'))
        mol.wrap()

        boxsize = mol.coords.max(axis=0) - mol.coords.min(axis=0)

        assert np.abs(mol.box - boxsize).max() < 2


if __name__ == "__main__":
    import unittest

    unittest.main(verbosity=2)


import os
import unittest
from tempfile import NamedTemporaryFile
from glob import glob
from moleculekit.home import home
from moleculekit.smallmol.smallmol import SmallMol
from moleculekit.smallmol.smallmollib import SmallMolLib
from pandas import core


SDF_N_MOLS = 100

SDF_IDS_DELETE = [1,10,15,16]
SDF_MOLNAME_DELETE = ['ZINC02583363', 'ZINC86860147', 'ZINC04342657', 'ZINC02023420']
SDF_FIELDS = ['ligname', '_mol']
SDF_LOC_0_99 = 'ZINC02141008'

class _TestSmallMol(unittest.TestCase):

    def setUp(self):
        self.dataDir= home('test-smallmol')

    def test_loaSdffile(self):
        sdffile = os.path.join(self.dataDir, 'fda_drugs_light.sdf')
        lib = SmallMolLib(sdffile)
        n_mols = lib.numMols
        self.assertEqual(n_mols, SDF_N_MOLS, 'Molecules not correctly loaded. '
                                             'Expected: {}; Now: {}'.format(SDF_N_MOLS, n_mols))

        sdffile = os.path.join(self.dataDir, 'fda_drugs_light.sdf.gz')
        lib = SmallMolLib(sdffile)
        n_mols = lib.numMols
        self.assertEqual(n_mols, SDF_N_MOLS, 'Molecules not correctly loaded. '
                                             'Expected: {}; Now: {}'.format(SDF_N_MOLS, n_mols))

    def test_writeSdf(self):
        sdffile = os.path.join(self.dataDir, 'fda_drugs_light.sdf')
        lib = SmallMolLib(sdffile)

        sdfname = NamedTemporaryFile().name + '.sdf'
        lib.writeSdf(sdfname)

        sdf_exists = os.path.isfile(sdfname)

        self.assertTrue(sdf_exists, msg="The sdf written was not found")

        sdf = SmallMolLib(sdfname)

        self.assertIsInstance(sdf, SmallMolLib, msg="The sdf written was not correctly loaded. Probably the previous"
                                                    "writing went wrong")


    def test_appendSmallMolLib(self):
        sdffile = os.path.join(self.dataDir, 'fda_drugs_light.sdf')
        lib = SmallMolLib(sdffile)
        lib2 = SmallMolLib(sdffile)

        lib.appendSmallLib(lib2)

        n_mol2_merged = lib.numMols

        self.assertEqual(n_mol2_merged, SDF_N_MOLS*2, msg="The number of molecules in the SmallMolLib is not as expected."
                                                          "The two sdf were not correctly merged. ")


    def test_appendSmallMol(self):
        mol2file = os.path.join(self.dataDir, 'benzamidine.mol2')
        sdffile = os.path.join(self.dataDir, 'fda_drugs_light.sdf')

        lib = SmallMolLib(sdffile)
        sm = SmallMol(mol2file)
        lib.appendSmallMol(sm)

        n_mol2_append = lib.numMols

        self.assertEqual(n_mol2_append, SDF_N_MOLS+1, msg="The number of molecules in the SmallMolLib is not as expected."
                                                          "The mol2 were not correctly append. ")

    def test_removeMols(self):
        sdffile = os.path.join(self.dataDir, 'fda_drugs_light.sdf')

        lib = SmallMolLib(sdffile)

        mols_ids = SDF_IDS_DELETE
        ref_mols_name = SDF_MOLNAME_DELETE

        mols_name = [ s.ligname for s in lib.getMols(mols_ids) ]

        self.assertListEqual(mols_name, ref_mols_name, msg="The molecules at the given indexes do not match with the"
                                                           "expected")
        lib.removeMols(mols_ids)

        mols_name_now = [ s.ligname for s in lib.getMols(mols_ids) ]

        self.assertFalse(mols_name_now == mols_name, msg="The molecules seem to not be deleted correctly")

    def test_convertToDataFrame(self):
        sdffile = os.path.join(self.dataDir, 'fda_drugs_light.sdf')

        lib = SmallMolLib(sdffile)

        df = lib.toDataFrame()

        self.assertIsInstance(df, core.frame.DataFrame, msg="The SmallMolLib object was not correctly converted into pandas"
                                                      "DataFrame")

        cols = df.columns.tolist()
        ref_cols =SDF_FIELDS

        self.assertEqual(cols, ref_cols, msg="The fields in the SmallMolLib object was not the expected one")

        ligname_99 = df.iloc[99][0]
        ref_ligname = SDF_LOC_0_99

        self.assertEqual(ligname_99, ref_ligname, msg="The ligand name found is not the expected one")


    def test_depict(self):
        import IPython
        refimg = os.path.join(self.dataDir, 'sdf.svg')
        sdffile = os.path.join(self.dataDir, 'fda_drugs_light.sdf')

        lib = SmallMolLib(sdffile)

        img_name = NamedTemporaryFile().name + '.svg'
        lib.depict(sketch=True, filename=img_name)
        _img = lib.depict(sketch=True, ipython=True)

        refimg_size = os.path.getsize(refimg)
        sm_img_size = os.path.getsize(img_name)

        self.assertIsInstance(_img, IPython.core.display.SVG, msg="The object is not an IPython image as expected")
        self.assertEqual(sm_img_size, refimg_size, msg="The svg image does not have the same size of the reference")


    def test_writeSmiles(self):
        sdffile = os.path.join(self.dataDir, 'fda_drugs_light.sdf')

        lib = SmallMolLib(sdffile)
        tmpfile = NamedTemporaryFile().name + '.smi'
        lib.writeSmiles(tmpfile)

        with open(tmpfile, 'r') as f:
            filelines = f.readlines()[1:]

        with open(os.path.join(self.dataDir, 'fda_drugs_light.smi'), 'r') as f:
            reflines = f.readlines()[1:]

        self.assertEqual(filelines, reflines)

    def test_readSmiles(self):
        smifile = os.path.join(self.dataDir, 'fda_drugs_light.smi')
        lib = SmallMolLib(smifile)
        assert len(lib) == 100

        smifile = os.path.join(self.dataDir, 'fda_drugs_light.smi.gz')
        lib = SmallMolLib(smifile)
        assert len(lib) == 100



if __name__ == '__main__':
    unittest.main(verbosity=2)


from rdkit import Chem
from rdkit.Chem import AllChem
import numpy as np
import logging
from moleculekit.smallmol.util import  convertToString
from tqdm import tqdm
from joblib import Parallel, delayed  # Import delayed as well for other modules


logger = logging.getLogger(__name__)


def ParallelExecutor(**joblib_args):
    """
    A wrapper for joblib.Parallel to allow custom progress bars.
    """
    def aprun(**tq_args):
        tqdm_f = lambda x, args: tqdm(x, **args)
        return lambda x: Parallel(**joblib_args)(tqdm_f(x, tq_args))
    return aprun



def getMaximumCommonSubstructure(smallmol_list, removeHs=True, returnAtomIdxs=False):
    """
    Returns the maximum common substructure and two list of lists. The first one contains for each molecules the atom
    indexes that are part of the MCS, the second list contains the indexes that are not part of the MCS.

    Parameters
    ----------
    smallmol_list: list
        The list of SmallMol objects
    removeHs: bool
        If True, the atom the hydrogens where not considered
        Default: True
    returnAtomIdxs: bool
        If True, the lists of the atom indexes are returned
        Default: False

    Returns
    -------
    mcs_mol: rdkit.Chem.rdchem.Mol
        The MCS molecule
    atom_mcs_list: list
        A list of lists containing the atom indexes that are part of the MCS
    atom_no_mcs_list: list
        A list of lists containing the atom indexes that are not part of the MCS
    """
    from rdkit.Chem import rdFMCS

    smallmol_list = [sm.copy() for sm in smallmol_list]
    for sm in smallmol_list:
        AllChem.EmbedMolecule(sm._mol, AllChem.ETKDG())
        if removeHs:
            sm._mol = Chem.RemoveHs(sm._mol)
    rdkitMols_list = [sm._mol for sm in smallmol_list]

    mcs = rdFMCS.FindMCS(rdkitMols_list)

    logger.info('MCS found a substructure of {} atoms and {} bonds'.format(mcs.numAtoms, mcs.numBonds))

    mcs_mol = Chem.MolFromSmarts(mcs.smartsString)

    if not returnAtomIdxs:
        return mcs_mol

    atoms_mcs_list = []
    atoms_no_mcs_list = []
    for sm, m in zip(smallmol_list, rdkitMols_list):
        match = m.GetSubstructMatch(mcs_mol)
        sel_str = convertToString(match)

        atoms_mcs = sm.get('idx', 'idx {}'.format(sel_str))
        atoms_no_mcs = sm.get('idx', 'idx {}'.format(sel_str),  invert=True)

        atoms_mcs_list.append(atoms_mcs.tolist())
        atoms_no_mcs_list.append(atoms_no_mcs.tolist())

    return mcs_mol, atoms_mcs_list, atoms_no_mcs_list

def cluster(smallmol_list, method, distThresholds=0.2, returnDetails=True, removeHs=True ):
    """
    Rreturn the SmallMol objects grouped in the cluster. It can also return the details of the clusters computed.

    Parameters
    ----------
    smallmol_list: list
        The list of moleculekit.smallmol.smallmol.SmallMol objects
    method: str
        The cluster methods. Can be  ['maccs', 'pathFingerprints', 'atomsFingerprints', 'torsionsFingerprints',
        'circularFingerprints', 'shape', 'mcs']
    distThresholds: float
        The disance cutoff for the clusters
        Default: 0.2
    returnDetails: bool
        If True, the cluster details are also returned
        Default: True
    removeHs: bool
        If True, the hydrogens are not considered
        Default: True

    Returns
    -------
    clusters: list
        List of lists, That contains the SmallMol objects grouped based on the cluster belongings
    details: list
        A list with all the cluster details
    """


    from sklearn.cluster import DBSCAN

    import sys
    this_module = sys.modules[__name__]

    _methods = ['maccs', 'pathFingerprints', 'atomsFingerprints', 'torsionsFingerprints',
                'circularFingerprints', 'shape', 'mcs']

    if method not in _methods:
        raise ValueError('The method provided {} does not exists. The ones available are the following: {}'.format(method, _methods))

    smallmol_list = np.array([sm.copy() for sm in smallmol_list])
    for sm in smallmol_list:
        AllChem.EmbedMolecule(sm._mol, AllChem.ETKDG())
        if removeHs:
            sm._mol = Chem.RemoveHs(sm._mol)
    rdkitMols_list = [sm._mol for sm in smallmol_list]

    clustmethod = getattr(this_module, '_{}Clustering'.format(method))

    if method not in ['shape', 'mcs']:
        matrix = clustmethod(rdkitMols_list)

    else:
        aprun = ParallelExecutor(n_jobs=-1)  # _config['ncpus'])
        matrix = aprun(total=len(rdkitMols_list), desc='{} Distance'.format(method)) \
                            (delayed(clustmethod)(mol1, rdkitMols_list) for mol1 in rdkitMols_list)

        matrix = np.array(matrix)


    db = DBSCAN(eps=distThresholds, min_samples=0, metric='precomputed' ).fit(matrix)

    labels = db.labels_

    populations = np.bincount(labels)
    n_clusters = np.max(labels)

    clusters_idx = np.empty((n_clusters,), dtype=object)
    clusters_smallmols = np.empty((n_clusters,), dtype=object)

    for n_cl in np.arange(n_clusters):
        idxs = np.where(labels == n_cl)[0]
        clusters_idx[n_cl] = idxs
        clusters_smallmols[n_cl] = smallmol_list[idxs]


    if returnDetails:
        details = {'numClusters':n_clusters,
                   'populations':populations,
                   'clusters':clusters_idx}
        return clusters_smallmols, details

    return clusters_smallmols




def _maccsClustering( rdkit_mols):
    """
    Returns the tanimoto distance matrix based on maccs method

    Parameters
    ----------
    rdkit_mols: list
        The list of rdkit.Chem.rdchem.Mol objects

    Returns
    -------
    tanimotomatrix: np.array
        The numpy array containing the tanimoto matrix
    """
    from rdkit.Chem import MACCSkeys  # calcola MACCS keys

    fps = []
    for m in tqdm(rdkit_mols):
       fps.append(MACCSkeys.GenMACCSKeys(m))

    aprun = ParallelExecutor(n_jobs=-1)  # _config['ncpus'])
    tanimoto_matrix = aprun(total=len(fps), desc='MACCS Distance') \
            (delayed(TanimotoDistances)(fp1, fps) for fp1 in fps)


    return np.array(tanimoto_matrix)

def _pathFingerprintsClustering(rdkit_mols):
    """
        Returns the tanimoto distance matrix based on fingerprints method

        Parameters
        ----------
        rdkit_mols: list
            The list of rdkit.Chem.rdchem.Mol objects

        Returns
        -------
        tanimotomatrix: np.array
            The numpy array containing the tanimoto matrix
        """
    from rdkit.Chem.Fingerprints import FingerprintMols  # calcola path fingerprints

    fps = [ ]
    for m in tqdm(rdkit_mols):
        fps.append(FingerprintMols.FingerprintMol(m))

    aprun = ParallelExecutor(n_jobs=-1)  # _config['ncpus'])
    tanimoto_matrix = aprun(total=len(fps), desc='PathFingerprints Distance') \
        (delayed(TanimotoDistances)(fp1, fps) for fp1 in fps)

    return np.array(tanimoto_matrix)

def _atomsFingerprintsClustering(rdkit_mols):
    """
        Returns the dice distance matrix based on atomsfingerprints method

        Parameters
        ----------
        rdkit_mols: list
            The list of rdkit.Chem.rdchem.Mol objects

        Returns
        -------
        dicematrix: np.array
            The numpy array containing the dice matrix
        """
    from rdkit.Chem.AtomPairs import Pairs  # Atom pairs

    fps = []
    for m in tqdm(rdkit_mols):
        fps.append(Pairs.GetAtomPairFingerprint(m))

    aprun = ParallelExecutor(n_jobs=-1)  # _config['ncpus'])
    dice_matrix = aprun(total=len(fps), desc='AtomsFingerprints Distance') \
        (delayed(DiceDistances)(fp1, fps) for fp1 in fps)

    return np.array(dice_matrix)

def _torsionsFingerprintsClustering(rdkit_mols):
    """
        Returns the dice distance matrix based on torsionsfingerprints method

        Parameters
        ----------
        rdkit_mols: list
            The list of rdkit.Chem.rdchem.Mol objects

        Returns
        -------
        dicematrix: np.array
            The numpy array containing the dice matrix
        """
    from rdkit.Chem.AtomPairs import Torsions  # Topological Torsions

    fps = [ ]
    for m in tqdm(rdkit_mols):
        fps.append(Torsions.GetHashedTopologicalTorsionFingerprint(m))

    aprun = ParallelExecutor(n_jobs=-1)  # _config['ncpus'])
    dice_matrix = aprun(total=len(fps), desc='TorsionsFingerprints Distance') \
        (delayed(DiceDistances)(fp1, fps) for fp1 in fps)

    return np.array(dice_matrix)

def _circularFingerprintsClustering(rdkit_mols, radius=2):
    """
        Returns the dice distance matrix based on circularfingerprints method

        Parameters
        ----------
        rdkit_mols: list
            The list of rdkit.Chem.rdchem.Mol objects

        radius: int
            The radius of the MorganCircularFingerprint
            Default: 2

        Returns
        -------
        dicematrix: np.array
            The numpy array containing the dice matrix
        """
    from rdkit.Chem import AllChem  # calcola circular fingerprints

    fps = []
    for m in rdkit_mols:
        fps.append(AllChem.GetMorganFingerprint(m, radius))

    aprun = ParallelExecutor(n_jobs=-1)  # _config['ncpus'])
    dice_matrix = aprun(total=len(fps), desc='CircularFingerprints Distance') \
        (delayed(DiceDistances)(fp1, fps) for fp1 in fps)

    return np.array(dice_matrix)

def _shapeClustering(mol1, rdkit_mols):
    """
        Returns the tanimoto row based on shape method

        Parameters
        ----------
        mol1: rdkit.Chem.rdchem.Mol
            The reference molecule
        rdkit_mols: list
            The list of rdkit.Chem.rdchem.Mol objects

        Returns
        -------
        tanimotorow: np.array
            The numpy array containing the tanimoto row
        """
    from rdkit.Chem import rdMolAlign, rdShapeHelpers
    tanimoto_shape_row = []
    for mol2 in rdkit_mols:
        oa3 = rdMolAlign.GetO3A(mol1, mol2)
        oa3.Align()
        tani_shape = rdShapeHelpers.ShapeTanimotoDist(mol1, mol2)
        tanimoto_shape_row.append(tani_shape)
    return tanimoto_shape_row

def _mcsClustering(mol1, rdkit_mols):
    """
       Returns the weight distance baased on mcs method

       Parameters
       ----------
       mol1: rdkit.Chem.rdchem.Mol
           The reference molecule
       rdkit_mols: list
           The list of rdkit.Chem.rdchem.Mol objects

       Returns
       -------
       distancerow: np.array
           The numpy array containing the distance row
           """
    from rdkit.Chem import rdFMCS

    MCS_row = []
    for mol2 in tqdm(rdkit_mols):
        sum_numHeavyAtoms = mol1.GetNumHeavyAtoms() + mol2.GetNumHeavyAtoms()
        mcsHeavyAtoms = rdFMCS.FindMCS([mol1, mol2], ringMatchesRingOnly=True, completeRingsOnly=True, timeout=5)
        mcsNumHeavyAtoms = float(Chem.MolFromSmarts(mcsHeavyAtoms.smartsString).GetNumHeavyAtoms())
        distance = 1 - mcsNumHeavyAtoms * 2/sum_numHeavyAtoms
        MCS_row.append(distance)

    return MCS_row

def TanimotoDistances(fp1, fps):
    """
    Returns the tanimoto row based on fingeprints passed

    Parameters
    ----------
    fp1: rdkit fingerprint
        The rdkit fingerprint computed used as reference
    fps: list
        The list of the rdkit fingerprint computed

    Returns
    -------
    tanimotorow: list
        A list with the tanimoto row similarities
    """

    from rdkit import DataStructs  # fingerprint similarity

    tanimoto_row = []

    for fp2 in fps:
        tani = 1 - DataStructs.FingerprintSimilarity(fp1, fp2)
        tanimoto_row.append(tani)
    return tanimoto_row


def DiceDistances(fp1, fps):
    """
        Returns the dice row based on fingeprints passed

        Parameters
        ----------
        fp1: rdkit fingerprint
            The rdkit fingerprint computed used as reference
        fps: list
            The list of the rdkit fingerprint computed

        Returns
        -------
        dicerow: list
            A list with the dice row similarities
        """
    from rdkit import DataStructs  # fingerprint similarity

    dice_row = []

    for fp2 in fps:
        dice = 1 - DataStructs.DiceSimilarity(fp1, fp2)
        dice_row.append(dice)

    return dice_row



# (c) 2015-2018 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#
from moleculekit.projections.projection import Projection
import numpy as np
import logging
logger = logging.getLogger(__name__)


class MetricSasa(Projection):
    """ Calculate solvent accessible surface area of a molecule.

    Implementation and documentation taken from MDtraj shrake_rupley code.

    Parameters
    ----------
    sel : str
        Atom selection string for atoms or residues for which to calculate the SASA.
        See more `here <http://www.ks.uiuc.edu/Research/vmd/vmd-1.9.2/ug/node89.html>`__
    probeRadius : float
        The radius of the probe, in Angstrom.
    numSpherePoints : int
        The number of points representing the surface of each atom, higher values lead to more accuracy.
    mode : str
        In mode == 'atom', the extracted areas are resolved per-atom. In mode == 'residue', this is consolidated down
        to the per-residue SASA by summing over the atoms in each residue.

    Returns
    -------
    metr : MetricSasa object
    """
    def __init__(self, sel='protein', probeRadius=1.4, numSpherePoints=960, mode='atom'):
        super().__init__()

        self._probeRadius = probeRadius
        self._numSpherePoints = numSpherePoints
        self._mode = mode
        self._sel = sel

    def _calculateMolProp(self, mol, props='all'):
        props = ('radii', 'atom_mapping', 'sel') if props == 'all' else props
        res = {}

        sel = mol.atomselect(self._sel)
        if 'sel' in props:
            res['sel'] = sel

        if 'radii' in props:
            _ATOMIC_RADII = {'C': 1.5, 'F': 1.2, 'H': 0.4, 'N': 1.10, 'O': 1.05, 'S': 1.6, 'P': 1.6}
            elements = [n[0] for n in mol.name[sel]]
            atom_radii = np.vectorize(_ATOMIC_RADII.__getitem__)(elements)
            res['radii'] = np.array(atom_radii, np.float32) + self._probeRadius

        if 'atom_mapping' in props:
            if self._mode == 'atom':
                res['atom_mapping'] = np.arange(np.sum(sel), dtype=np.int32)
            elif self._mode == 'residue':
                from moleculekit.util import sequenceID
                res['atom_mapping'] = sequenceID((mol.resid[sel], mol.chain[sel], mol.segid[sel])).astype(np.int32)
            else:
                raise ValueError('mode must be one of "residue", "atom". "{}" supplied'.format(self._mode))

        return res

    def project(self, mol):
        """ Project molecule.

        Parameters
        ----------
        mol : :class:`Molecule <moleculekit.molecule.Molecule>`
            A :class:`Molecule <moleculekit.molecule.Molecule>` object to project.

        Returns
        -------
        data : np.ndarray
            An array containing the projected data.
        """
        getMolProp = lambda prop: self._getMolProp(mol, prop)
        radii = getMolProp('radii')
        atom_mapping = getMolProp('atom_mapping')
        sel = getMolProp('sel')

        xyz = np.swapaxes(np.swapaxes(np.atleast_3d(mol.coords[sel, :, :]), 1, 2), 0, 1)
        xyz = np.array(xyz.copy(), dtype=np.float32) / 10  # converting to nm

        try:
            from mdtraj.geometry._geometry import _sasa as sasa
        except ImportError:
            raise ImportError('To calculate SASA you need to install mdtraj with `conda install mdtraj -c omnia`')

        out = np.zeros((mol.numFrames, atom_mapping.max() + 1), dtype=np.float32)
        sasa(xyz, radii / 10, int(self._numSpherePoints), atom_mapping, out)  # Divide radii by 10 for nm
        return out

    def getMapping(self, mol):
        """ Returns the description of each projected dimension.

        Parameters
        ----------
        mol : :class:`Molecule <moleculekit.molecule.Molecule>` object
            A Molecule object which will be used to calculate the descriptions of the projected dimensions.

        Returns
        -------
        map : :class:`DataFrame <pandas.core.frame.DataFrame>` object
            A DataFrame containing the descriptions of each dimension
        """
        getMolProp = lambda prop: self._getMolProp(mol, prop)
        atom_mapping = getMolProp('atom_mapping')
        atomsel = getMolProp('sel')

        if self._mode == 'atom':
            atomidx = np.where(atomsel)[0]
        elif self._mode == 'residue':
            _, firstidx = np.unique(atom_mapping, return_index=True)
            atomidx = np.where(atomsel)[0][firstidx]
        else:
            raise ValueError('mode must be one of "residue", "atom". "{}" supplied'.format(self._mode))

        from pandas import DataFrame
        types = []
        indexes = []
        description = []
        for i in atomidx:
            types += ['SASA']
            indexes += [i]
            description += ['SASA of {} {} {}'.format(mol.resname[i], mol.resid[i], mol.name[i])]
        return DataFrame({'type': types, 'atomIndexes': indexes, 'description': description})


import unittest
class _TestMetricSasa(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        from moleculekit.molecule import Molecule
        from moleculekit.home import home
        from os import path
        
        mol = Molecule(path.join(home(dataDir='test-projections'), 'trajectory', 'filtered.pdb'))
        mol.read(path.join(home(dataDir='test-projections'), 'trajectory', 'traj.xtc'))
        mol.dropFrames(keep=[0, 1])  # Keep only two frames because it's super slow
        self.mol = mol

    def test_sasa_atom(self):
        from os import path
        from moleculekit.home import home

        metr = MetricSasa(mode='atom')
        sasaA = metr.project(self.mol.copy())
        sasaA_ref = np.load(path.join(home(dataDir='test-projections'), 'metricsasa', 'sasa_atom.npy'))
        assert np.allclose(sasaA, sasaA_ref, atol=7e-4)

    def test_sasa_residue(self):
        from os import path
        from moleculekit.home import home

        metr = MetricSasa(mode='residue')
        sasaR = metr.project(self.mol.copy())
        sasaR_ref = np.load(path.join(home(dataDir='test-projections'), 'metricsasa', 'sasa_residue.npy'))
        assert np.allclose(sasaR, sasaR_ref, atol=3e-3)


if __name__ == '__main__':
    unittest.main(verbosity=2)


# (c) 2015-2018 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#
import numpy as np
import logging
import moleculekit.molecule

logger = logging.getLogger(__name__)


def pp_calcDistances(mol, sel1, sel2, metric='distances', threshold=8, pbc=True, gap=1, truncate=None):
    distances = _distanceArray(mol, sel1, sel2, pbc)
    distances = _postProcessDistances(distances, sel1, sel2, truncate)

    if metric == 'contacts':
        # from scipy.sparse import lil_matrix
        # metric = lil_matrix(distances <= threshold)
        metric = distances <= threshold
    elif metric == 'distances':
        metric = distances.astype(dtype=np.float32)
    else:
        raise NameError('The metric you asked for is not supported. Check spelling and documentation')
    return metric


def pp_calcMinDistances(mol, sel1, sel2, metric='distances', threshold=8, pbc=True, gap=1, truncate=None):
    import os
    import ctypes
    from moleculekit.home import home

    # Converting non-grouped boolean atomselection to group-style atomselections
    if np.ndim(sel1) != 2:
        sel1idx = tuple(np.where(sel1)[0])
        sel1 = np.zeros((len(sel1idx), len(sel1)), dtype=bool)
        sel1[range(sel1.shape[0]), sel1idx] = True
    if np.ndim(sel2) != 2:
        sel2idx = tuple(np.where(sel2)[0])
        sel2 = np.zeros((len(sel2idx), len(sel2)), dtype=bool)
        sel2[range(sel2.shape[0]), sel2idx] = True

    coords = mol.coords
    box = mol.box
    if pbc:
        if box is None or np.sum(box) == 0:
            raise RuntimeError('No periodic box dimensions given in the molecule/trajectory. '
                            'If you want to calculate distance without wrapping, set the pbc option to False')
    else:
        box = np.zeros((3, coords.shape[2]), dtype=np.float32)

    if box.shape[1] != coords.shape[2]:
        raise RuntimeError('Different number of frames in mol.coords and mol.box. '
                            'Please ensure they both have the same number of frames')

    # Converting from 2D boolean atomselect array to 2D int array where each row starts with the indexes of the boolean
    groups1 = np.ones((sel1.shape[0], mol.numAtoms), dtype=np.int32) * -1
    groups2 = np.ones((sel2.shape[0], mol.numAtoms), dtype=np.int32) * -1
    for i in range(sel1.shape[0]):
        idx = np.where(sel1[i, :])[0]
        groups1[i, 0:len(idx)] = idx
    for i in range(sel2.shape[0]):
        idx = np.where(sel2[i, :])[0]
        groups2[i, 0:len(idx)] = idx

    selfdist = np.array_equal(sel1, sel2)

    # Running the actual calculations
    lib = ctypes.cdll.LoadLibrary(os.path.join(home(libDir=True), 'mindist_ext.so'))
    mindist = np.zeros((mol.numFrames, len(groups1) * len(groups2)), dtype=np.float32)  # Preparing the return array
    if selfdist:
        mindist = np.zeros((mol.numFrames, int((len(groups1) * (len(groups2)-1))/2)), dtype=np.float32)

    #import time
    #t = time.time()
    lib.mindist_trajectory(coords.ctypes.data_as(ctypes.POINTER(ctypes.c_float)),
                           box.ctypes.data_as(ctypes.POINTER(ctypes.c_float)),
                           groups1.ctypes.data_as(ctypes.POINTER(ctypes.c_int)),
                           groups2.ctypes.data_as(ctypes.POINTER(ctypes.c_int)),
                           ctypes.c_int(len(groups1)),
                           ctypes.c_int(len(groups2)),
                           ctypes.c_int(mol.numAtoms),
                           ctypes.c_int(mol.numFrames),
                           ctypes.c_int(int(pbc)),
                           ctypes.c_int(int(selfdist)),
                           mindist.ctypes.data_as(ctypes.POINTER(ctypes.c_float)))
    #print(time.time() - t)

    if truncate is not None:
        mindist[mindist > truncate] = truncate

    if metric == 'contacts':
        mindist = mindist <= threshold
    elif metric == 'distances':
        mindist = mindist.astype(dtype=np.float32)
    else:
        raise NameError('The metric you asked for is not supported. Check spelling and documentation')
    return mindist


def _postProcessDistances(distances, sel1, sel2, truncate):
    # distances is a list of numpy arrays. Each numpy array is numFrames x numSel1. The list is length numSel2
    # Setting upper triangle to -1 if same selections
    if np.array_equal(sel1, sel2):
        for i in range(len(distances)):
                distances[i][:, range(i + 1)] = -1

    if np.ndim(distances[0]) > 1:  # 2D data
        distances = np.concatenate(distances, axis=1)
    else:  # 1D data
        distances = np.vstack(distances).transpose()

    if np.array_equal(sel1, sel2):
        distances = distances[:, np.all(distances != -1, 0)]

    if truncate is not None:
        distances[distances > truncate] = truncate
    return np.atleast_1d(np.squeeze(distances))


def _distanceArray(mol, sel1, sel2, pbc):
    numsel1 = np.sum(sel1)
    numsel2 = np.sum(sel2)
    coords1 = mol.coords[sel1, :, :]
    coords2 = mol.coords[sel2, :, :]

    distances = []
    for j in range(numsel2):
        coo2 = coords2[j, :, :]  # 3 x numframes array
        dists = coords1 - coo2
        if pbc:
            if mol.box is None or np.sum(mol.box) == 0:
                raise NameError(
                    'No periodic box dimensions given in the molecule/trajectory. If you want to calculate distance without wrapping, set the pbc option to False')
            dists = _wrapDistances(mol.box, dists, _findDiffChain(mol, sel1, sel2, j, range(numsel1)))
        dists = np.transpose(np.sqrt(np.sum(dists * dists, 1)))
        distances.append(dists)
    return distances


def _findDiffChain(mol, sel1, sel2, i, others):
    if np.array_equal(sel1, sel2) and len(mol.chain) > 0:
        chain = mol.get('chain', sel=sel1)
        diffchain = chain[others] != chain[i]
    else:
        diffchain = None
    return diffchain


def _wrapDistances(box, dist, diffchain):
    if diffchain is not None:
        dist[diffchain, :, :] -= box * np.round(dist[diffchain, :, :] / box)
    else:
        dist = dist - box * np.round(dist / box)
    return dist




# (c) 2015-2018 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#

from moleculekit.projections.projection import Projection
from moleculekit.projections.util import pp_calcDistances, pp_calcMinDistances
from moleculekit.util import ensurelist
import numpy as np
import logging
logger = logging.getLogger(__name__)


class MetricDistance(Projection):
    """ Creates a MetricDistance object

    Parameters
    ----------
    sel1 : str
        Atom selection string for the first set of atoms.
        See more `here <http://www.ks.uiuc.edu/Research/vmd/vmd-1.9.2/ug/node89.html>`__
    sel2 : str
        Atom selection string for the second set of atoms. If sel1 != sel2, it will calculate inter-set distances.
        If sel1 == sel2, it will calculate intra-set distances.
        See more `here <http://www.ks.uiuc.edu/Research/vmd/vmd-1.9.2/ug/node89.html>`__
    groupsel1 : ['all','residue'], optional
        Group all atoms in `sel1` to the single minimum distance. Alternatively can calculate the minimum distance of a
        residue containing the atoms in sel1.
    groupsel2 : ['all','residue'], optional
        Same as groupsel1 but for `sel2`
    metric : ['distances','contacts'], optional
        Set to 'contacts' to calculate contacts instead of distances
    threshold : float, optional
        The threshold under which a distance is considered in contact
    pbc : bool, optional
        Set to false to disable distance calculations using periodic distances
    truncate : float, optional
        Set all distances larger than `truncate` to `truncate`
    update :
        Not functional yet

    Returns
    -------
    proj : MetricDistance object
    """
    def __init__(self, sel1, sel2, groupsel1=None, groupsel2=None,  metric='distances', threshold=8, pbc=True, truncate=None):
        super().__init__()

        self.sel1 = sel1
        self.sel2 = sel2
        self.groupsel1 = groupsel1
        self.groupsel2 = groupsel2
        self.metric = metric
        self.threshold = threshold
        self.pbc = pbc
        self.truncate = truncate

    def project(self, mol):
        """ Project molecule.

        Parameters
        ----------
        mol : :class:`Molecule <moleculekit.molecule.Molecule>`
            A :class:`Molecule <moleculekit.molecule.Molecule>` object to project.

        Returns
        -------
        data : np.ndarray
            An array containing the projected data.
        """
        getMolProp = lambda prop: self._getMolProp(mol, prop)
        sel1 = getMolProp('sel1')
        sel2 = getMolProp('sel2')

        if np.ndim(sel1) == 1 and np.ndim(sel2) == 1:  # normal distances
            metric = pp_calcDistances(mol, sel1, sel2, self.metric, self.threshold, self.pbc, truncate=self.truncate)
        else:  # minimum distances by groups
            metric = pp_calcMinDistances(mol, sel1, sel2, self.metric, self.threshold, pbc=self.pbc, truncate=self.truncate)

        return metric

    def _calculateMolProp(self, mol, props='all'):
        props = ('sel1', 'sel2') if props == 'all' else props
        res = {}
        if 'sel1' in props:
            res['sel1'] = self._processSelection(mol, self.sel1, self.groupsel1)
        if 'sel2' in props:
            res['sel2'] = self._processSelection(mol, self.sel2, self.groupsel2)
        return res

    def _processSelection(self, mol, sel, groupsel):
        if isinstance(sel, str):  # If user passes simple string selections
            if groupsel is None:
                sel = mol.atomselect(sel)
            elif groupsel == 'all':
                sel = self._processMultiSelections(mol, [sel])
            elif groupsel == 'residue':
                sel = self._groupByResidue(mol, sel)
            else:
                raise NameError('Invalid groupsel argument')
        else:  # If user passes his own sets of groups
            sel = self._processMultiSelections(mol, sel)

        if np.sum(sel) == 0:
            raise NameError('Selection returned 0 atoms')
        return sel

    def _processMultiSelections(self, mol, sel):
        newsel = np.zeros((len(sel), mol.numAtoms), dtype=bool)
        for s in range(len(sel)):
            newsel[s, :] = mol.atomselect(sel[s])
        return newsel

    def _groupByResidue(self, mol, sel):
        import pandas as pd
        idx = mol.atomselect(sel, indexes=True)
        df = pd.DataFrame({'a': mol.resid[idx]})
        gg = df.groupby(by=df.a).groups  # Grouping by same resids

        newsel = np.zeros((len(gg), mol.numAtoms), dtype=bool)
        for i, res in enumerate(sorted(gg)):
            newsel[i, idx[gg[res]]] = True  # Setting the selected indexes to True which correspond to the same residue
        return newsel

    def getMapping(self, mol):
        """ Returns the description of each projected dimension.

        Parameters
        ----------
        mol : :class:`Molecule <moleculekit.molecule.Molecule>` object
            A Molecule object which will be used to calculate the descriptions of the projected dimensions.

        Returns
        -------
        map : :class:`DataFrame <pandas.core.frame.DataFrame>` object
            A DataFrame containing the descriptions of each dimension
        """
        getMolProp = lambda prop: self._getMolProp(mol, prop)
        sel1 = getMolProp('sel1')
        sel2 = getMolProp('sel2')

        if np.ndim(sel1) == 2:
            protatoms = []
            for i in range(sel1.shape[0]):
                protatoms.append(np.where(sel1[i, :] == True)[0])
        else:
            protatoms = np.where(sel1)[0]
        if np.ndim(sel2) == 2:
            ligatoms = []
            for i in range(sel2.shape[0]):
                ligatoms.append(np.where(sel2[i, :] == True)[0])
        else:
            ligatoms = np.where(sel2)[0]

        numatoms1 = len(protatoms)
        numatoms2 = len(ligatoms)

        from pandas import DataFrame
        types = []
        indexes = []
        description = []
        if np.array_equal(sel1, sel2):
            for i in range(numatoms1):
                for j in range(i+1, numatoms1):
                    atm1 = protatoms[i]
                    atm2 = protatoms[j]
                    desc = '{} between {} {} {} and {} {} {}'.format(self.metric[:-1],
                                                                     mol.resname[atm1], mol.resid[atm1], mol.name[atm1],
                                                                     mol.resname[atm2], mol.resid[atm2], mol.name[atm2])
                    types += [self.metric[:-1]]
                    indexes += [[atm1, atm2]]
                    description += [desc]
        else:
            for j in range(numatoms2):
                for i in range(numatoms1):
                    atm1 = protatoms[i]
                    atm2 = ligatoms[j]
                    desc = '{} between {} {} {} and {} {} {}'.format(self.metric[:-1],
                                                                     mol.resname[atm1], mol.resid[atm1], mol.name[atm1],
                                                                     mol.resname[atm2], mol.resid[atm2], mol.name[atm2])
                    types += [self.metric[:-1]]
                    indexes += [[atm1, atm2]]
                    description += [desc]
        return DataFrame({'type': types, 'atomIndexes': indexes, 'description': description})


class MetricSelfDistance(MetricDistance):
    def __init__(self, sel, groupsel=None, metric='distances', threshold=8, pbc=True, truncate=None):
        """ Creates a MetricSelfDistance object

        Parameters
        ----------
        sel : str
            Atom selection string for which to calculate the self distance.
            See more `here <http://www.ks.uiuc.edu/Research/vmd/vmd-1.9.2/ug/node89.html>`__
        groupsel : ['all','residue'], optional
            Group all atoms in `sel` to the single minimum distance. Alternatively can calculate the minimum distance
            of a residue containing the atoms in `sel`.
        metric : ['distances','contacts'], optional
            Set to 'contacts' to calculate contacts instead of distances
        threshold : float, optional
            The threshold under which a distance is considered in contact
        pbc : bool, optional
            Set to false to disable distance calculations using periodic distances
        truncate : float, optional
            Set all distances larger than `truncate` to `truncate`
        update :
            Not functional yet

        Returns
        -------
        proj : MetricDistance object
        """
        super().__init__(sel, sel, groupsel, groupsel, metric, threshold, pbc, truncate)


def contactVecToMatrix(vector, atomIndexes):
    from copy import deepcopy
    # Calculating the unique atom groups in the mapping
    uqAtomGroups = []
    atomIndexes = deepcopy(list(atomIndexes))
    for ax in atomIndexes:
        ax[0] = ensurelist(ax[0])
        ax[1] = ensurelist(ax[1])
        if ax[0] not in uqAtomGroups:
            uqAtomGroups.append(ax[0])
        if ax[1] not in uqAtomGroups:
            uqAtomGroups.append(ax[1])
    uqAtomGroups.sort(key=lambda x: x[0])  # Sort by first atom in each atom list
    num = len(uqAtomGroups)

    matrix = np.zeros((num, num), dtype=vector.dtype)
    mapping = np.ones((num, num), dtype=int) * -1
    for i in range(len(vector)):
        row = uqAtomGroups.index(atomIndexes[i][0])
        col = uqAtomGroups.index(atomIndexes[i][1])
        matrix[row, col] = vector[i]
        matrix[col, row] = vector[i]
        mapping[row, col] = i
        mapping[col, row] = i
    return matrix, mapping, uqAtomGroups


def reconstructContactMap(vector, mapping, truecontacts=None, plot=True, figsize=(7, 7), dpi=80, title=None, outfile=None, colors=None):
    """ Plots a given vector as a contact map

    Parameters
    ----------
    vector : np.ndarray or list
        A 1D vector of contacts
    mapping : pd.DataFrame
        A pandas DataFrame which describes the dimensions of the projection
    truecontacts : np.ndarray or list
        A 1D vector of true contacts
    plot : bool
        To plot or not to plot
    figsize : tuple
        The size of the final plot in inches
    dpi : int
        Dots per inch
    outfile : str
        Path of file in which to save the plot

    Returns
    -------
    cm : np.ndarray
        The input vector converted into a 2D numpy array

    Examples
    --------
    >>> reconstructContactMap(contacts, mapping)
    To use it with distances instead of contacts pass ones as the concat vector
    >>> reconstructContactMap(np.ones(dists.shape, dtype=bool), mapping, colors=dists)
    """

    from copy import deepcopy
    from matplotlib import cm as colormaps
    if truecontacts is None:
        truecontacts = np.zeros(len(vector), dtype=bool)
    if len(vector) != len(mapping):
        raise RuntimeError('Vector and map length must match.')

    # Checking if contacts or distances exist in the data
    contactidx = mapping.type == 'contact'
    if not np.any(contactidx):
        contactidx = mapping.type == 'distance'
        if not np.any(contactidx):
            raise RuntimeError(
                'No contacts or distances found in the MetricData object. Check the `.map` property of the object for a description of your projection.')

    # Creating the 2D contact maps
    cm, newmapping, uqAtomGroups = contactVecToMatrix(vector, mapping.atomIndexes)
    cmtrue, _, _ = contactVecToMatrix(truecontacts, mapping.atomIndexes)
    num = len(uqAtomGroups)

    if plot:
        from matplotlib import pylab as plt
        f = plt.figure(figsize=figsize, dpi=dpi)
        plt.imshow(cmtrue / 2, interpolation='none', vmin=0, vmax=1, aspect='equal',
                   cmap='Greys')  # /2 to convert to gray from black

        rows, cols = np.where(cm)
        colorbar = False
        if colors is None:
            truecms = vector & truecontacts
            colors = np.array(['r']*len(vector), dtype=object)
            colors[truecms] = '#ffff00'
        elif isinstance(colors, np.ndarray) and isinstance(colors[0], float):
            mpbl = colormaps.ScalarMappable(cmap=colormaps.jet)
            mpbl.set_array(colors)
            colors = mpbl.to_rgba(colors)
            colorbar = True
        if len(colors) == len(vector):
            colors = colors[newmapping[rows, cols]]

        plt.scatter(rows, cols, s=figsize[0] * 5, marker='o', c=colors, lw=0)
        if colorbar:
            plt.colorbar(mpbl)

        ax = f.axes[0]
        # Major ticks
        ax.set_xticks(np.arange(0, num, 1))
        ax.set_yticks(np.arange(0, num, 1))

        # Labels for major ticks
        ax.set_xticklabels([x[0] for x in uqAtomGroups])
        ax.set_yticklabels([x[0] for x in uqAtomGroups], )

        # Minor ticks
        ax.set_xticks(np.arange(-.5, num, 1), minor=True)
        ax.set_yticks(np.arange(-.5, num, 1), minor=True)
        plt.setp(ax.xaxis.get_majorticklabels(), rotation=90)

        # Gridlines based on minor ticks
        ax.grid(which='minor', color='#969696', linestyle='-', linewidth=1)
        ax.tick_params(axis='both', which='both', length=0)
        plt.xlim([-.5, num - .5])
        plt.ylim([-.5, num - .5])
        plt.xlabel('Atom index')
        plt.ylabel('Atom index')
        if title:
            plt.title(title)
        if outfile is not None:
            plt.savefig(outfile, dpi=dpi, bbox_inches='tight', pad_inches=0.2)
            plt.close()
        else:
            plt.show()
    return cm

import unittest
from moleculekit.home import home
import os
class _TestMetricDistance(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        from moleculekit.molecule import Molecule
        from os import path
        self.mol = Molecule(path.join(home(dataDir='test-projections'), 'trajectory', 'filtered.pdb'))
        self.mol_skipped = self.mol.copy()

        self.mol.read(path.join(home(dataDir='test-projections'), 'trajectory', 'traj.xtc'))
        self.mol_skipped.read(path.join(home(dataDir='test-projections'), 'trajectory', 'traj.xtc'), skip=10)

    def test_contacts(self):
        metr = MetricDistance('protein and name CA', 'resname MOL and noh', metric='contacts')
        data = metr.project(self.mol)
        refdata = np.load(os.path.join(home(dataDir='test-projections'), 'metricdistance', 'contacts.npy'))
        assert np.allclose(data, refdata, atol=1e-3), 'Contact calculation is broken'

    def test_distances(self):
        metr = MetricDistance('protein and name CA', 'resname MOL and noh', metric='distances')
        data = metr.project(self.mol)
        refdata = np.load(os.path.join(home(dataDir='test-projections'), 'metricdistance', 'distances.npy'))
        assert np.allclose(data, refdata, atol=1e-3), 'Distance calculation is broken'

    def test_mindistances(self):
        metr = MetricDistance('protein and noh', 'resname MOL and noh', groupsel1='residue', groupsel2='all')
        data = metr.project(self.mol)
        refdata = np.load(os.path.join(home(dataDir='test-projections'), 'metricdistance', 'mindistances.npy'))
        assert np.allclose(data, refdata, atol=1e-3), 'Minimum distance calculation is broken'

    def test_mindistances_truncate(self):
        metr = MetricDistance('protein and noh', 'resname MOL and noh', groupsel1='residue', groupsel2='all', truncate=3)
        data = metr.project(self.mol)
        refdata = np.load(os.path.join(home(dataDir='test-projections'), 'metricdistance', 'mindistances.npy'))
        assert np.allclose(data, np.clip(refdata, 0, 3), atol=1e-3), 'Minimum distance calculation is broken'

    def test_selfmindistance_manual(self):
        metr = MetricDistance('protein and resid 1 to 50 and noh', 'protein and resid 1 to 50 and noh', groupsel1='residue', groupsel2='residue')
        data = metr.project(self.mol)
        refdata = np.load(os.path.join(home(dataDir='test-projections'), 'metricdistance', 'selfmindistance.npy'))
        assert np.allclose(data, refdata, atol=1e-3), 'Manual self-distance is broken'

    def test_selfmindistance_auto(self):
        metr = MetricSelfDistance('protein and resid 1 to 50 and noh', groupsel='residue')
        data = metr.project(self.mol)
        refdata = np.load(os.path.join(home(dataDir='test-projections'), 'metricdistance', 'selfmindistance.npy'))
        assert np.allclose(data, refdata, atol=1e-3), 'Automatic self-distance is broken'

    def test_mindistances_skip(self):
        metr = MetricSelfDistance('protein and resid 1 to 50 and noh', groupsel='residue')
        data = metr.project(self.mol_skipped)  
        refdata = np.load(os.path.join(home(dataDir='test-projections'), 'metricdistance', 'selfmindistance.npy'))
        assert np.allclose(data, refdata[::10, :], atol=1e-3), 'Minimum distance calculation with skipping is broken'

    def test_reconstruct_contact_map(self):
        from moleculekit.util import tempname
        from moleculekit.molecule import Molecule
        import matplotlib
        matplotlib.use('Agg')
        
        mol = Molecule('1yu8')

        metr = MetricSelfDistance('protein and name CA', metric='contacts', pbc=False)
        data = metr.project(mol)
        mapping = metr.getMapping(mol)

        tmpfile = tempname(suffix='.svg')
        cm, newmapping, uqAtomGroups = contactVecToMatrix(data, mapping.atomIndexes)
        reconstructContactMap(data, mapping, outfile=tmpfile)
        refdata = np.load(os.path.join(home(dataDir='test-projections'), 'metricdistance', 'contactvectomatrix.npz'))
        
        assert np.array_equal(refdata['cm'], cm)
        assert np.array_equal(refdata['newmapping'], newmapping)
        assert np.array_equal(refdata['uqAtomGroups'], uqAtomGroups)




if __name__ == '__main__':
    unittest.main(verbosity=2)



import os
import string
from tempfile import NamedTemporaryFile
import numpy as np
from moleculekit.tools.autosegment import autoSegment2
from moleculekit.molecule import Molecule
from moleculekit.writers import _deduce_PDB_atom_name, checkTruncations
from moleculekit.util import ensurelist
import logging


logger = logging.getLogger(__name__)


def getPDBQTAtomType(atype, aidx, mol, aromaticNitrogen=False):
    tmptype = ''
    # carbons
    if atype == 'Car':
        tmptype = 'A'
    elif atype.startswith('C'):
        tmptype = 'C'
    # nitrogens
    if atype.startswith('N'):
        tmptype = 'N'
        if atype in ['Nam', 'Npl', 'Ng+']:
            bs, bo = np.where(mol.bonds == aidx)
            if len(bs) == 2:
                tmptype += 'A'
        elif atype == 'Nar':
            # if mol.resname[aidx] == 'HIE':
            bs, bo = np.where(mol.bonds == aidx)
            if len(bs) == 2:
                tmptype += 'a' if aromaticNitrogen else 'A'
            else:
                tmptype += 'n' if aromaticNitrogen else ''
        elif atype[-1] != '+':
            tmptype += 'A'
    # elif atype.startswith('N'):
    #   print(atype, aidx)
    #  tmptype = 'NA'
    # oxygens
    if atype.startswith('O'):
        tmptype = 'OA'
    # sulfurs
    if atype.startswith('S'):
        tmptype = 'S'
        if atype not in ['Sox', 'Sac']:
            tmptype += 'A'

    # hydrogens
    if atype.startswith('H'):
        tmptype = 'H'
        # print(aidx)
        # print(np.where(mol.bonds == aidx))
        bond = np.where(mol.bonds == aidx)[0][0]
        oidx = [a for a in mol.bonds[bond] if a != aidx][0]
        if mol.element[oidx] not in ['C', 'A']:
            tmptype += 'D'
    if tmptype == '':
        tmptype = atype[0]

    return tmptype


def getProperties(mol):
    try:
        import pybel
    except ImportError:
        raise ImportError('Could not import openbabel. The atomtyper requires this dependency so please install it with `conda install openbabel -c acellera`')
    name = NamedTemporaryFile(suffix='.pdb').name
    mol.write(name)
    mpybel = next(pybel.readfile('pdb', name))
    # print(name)
    residues = pybel.ob.OBResidueIter(mpybel.OBMol)
    atoms = [[r.GetName(), r.GetNum(), r.GetAtomID(at), at.GetType(), round(at.GetPartialCharge(), 3)]
             for r in residues
             for at in pybel.ob.OBResidueAtomIter(r)]
    return atoms

def prepareProteinForAtomtyping(mol, guessBonds=True, protonate=True, pH=7, segment=True, verbose=True):
    """ Prepares a Molecule object for atom typing.

    Parameters
    ----------
    mol : Molecule object
        The protein to prepare
    guessBonds : bool
        Drops the bonds in the molecule and guesses them from scratch
    protonate : bool
        Protonates the protein for the given pH and optimizes hydrogen networks
    pH : float
        The pH for protonation
    segment : bool
        Automatically guesses the segments of a protein by using the guessed bonds
    verbose : bool
        Set to False to turn of the printing

    Returns
    -------
    mol : Molecule object
        The prepared Molecule
    """
    mol = mol.copy()
    protsel = mol.atomselect('protein')

    if not np.any(protsel):
        raise RuntimeError('No protein atoms found in Molecule')

    if np.any(~protsel):
        resnames = np.unique(mol.resname[~protsel])
        raise RuntimeError('Found non-protein atoms with resnames {} in the Molecule. Please make sure to only pass protein atoms.'.format(resnames))

    if protonate:
        from moleculekit.tools.preparation import proteinPrepare
        mol = proteinPrepare(mol, pH=pH, verbose=verbose, _loggerLevel='INFO' if verbose else 'ERROR')

    if guessBonds:
        mol.bonds = mol._guessBonds()

    if segment:
        from moleculekit.tools.autosegment import autoSegment2
        mol = autoSegment2(mol, fields=('segid', 'chain'), _logger=verbose) 
    return mol


def atomtypingValidityChecks(mol):
    logger.info('Checking validity of Molecule before atomtyping. ' \
                'If it gives incorrect results or to improve performance disable it with validitychecks=False. ' \
                'Most of these checks can be passed by using the moleculekit.atomtyper.prepareProteinForAtomtyping function. ' \
                'But make sure you understand what you are doing.')
    protsel = mol.atomselect('protein')

    if not np.any(protsel):
        raise RuntimeError('No protein atoms found in Molecule')

    if np.any(~protsel):
        resnames = np.unique(mol.resname[~protsel])
        raise RuntimeError('Found non-protein atoms with resnames {} in the Molecule. Please make sure to only pass protein atoms.'.format(resnames))

    if mol.bonds.shape[0] < mol.numAtoms:
        raise ValueError('The protein has less bonds than atoms. This seems incorrect. Assign them with `mol.bonds = mol._getBonds()`')

    if np.all(mol.segid == '') or np.all(mol.chain == ''):
        raise RuntimeError('Please assign segments to the segid and chain fields of the molecule using autoSegment2')

    from moleculekit.tools.autosegment import autoSegment2
    mm = mol.copy()
    mm.segid[:] = ''  # Set segid and chain to '' to avoid name clashes in autoSegment2
    mm.chain[:] = ''
    refmol = autoSegment2(mm, fields=('chain', 'segid'), _logger=False)
    numsegsref = len(np.unique(refmol.segid))
    numsegs = len(np.unique(mol.segid))
    if numsegs != numsegsref:
        raise RuntimeError('The molecule contains {} segments while we predict {}. Make sure you used autoSegment2 on the protein'.format(numsegs, numsegsref))

    if not np.any(mol.element == 'H'):
        raise RuntimeError('No hydrogens found in the Molecule. Make sure to use proteinPrepare before passing it to voxelization. Also you might need to recalculate the bonds after this.')


def getPDBQTAtomTypesAndCharges(mol, aromaticNitrogen=False, validitychecks=True):
    if validitychecks:
        atomtypingValidityChecks(mol)

    atomsProp = getProperties(mol)
    for n, a in enumerate(atomsProp):
        if a[0] == 'HIP':
            if a[2].strip().startswith('C') and a[2].strip() not in ['CA', 'C', 'CB']:
                a[3] = 'Car'

            #print(n, a)
    charges = ['{0:.3f}'.format(a[-1]) for a in atomsProp]
    pdbqtATypes = [getPDBQTAtomType(a[3], n, mol, aromaticNitrogen)
                   for n, a in enumerate(atomsProp)]

    return np.array(pdbqtATypes, dtype='O'), np.array(charges, dtype='float32')


def _getHydrophobic(atypes):
    return atypes == 'C'


def _getAromatic(atypes):
    return (atypes == 'A') | (atypes == 'Na') | (atypes == 'Nn')


def _getAcceptor(atypes):
    return (atypes == 'OA') | (atypes == 'NA') | (atypes == 'SA') | (atypes == 'Na')


def _getDonors(atypes, bonds):
    donors = np.zeros(len(atypes), dtype=bool)
    hydrogens = np.where((atypes == 'HD') | (atypes == 'HS'))[0]
    for h in hydrogens:
        partners = bonds[bonds[:, 0] == h, 1]
        partners = np.hstack((partners, bonds[bonds[:, 1] == h, 0]))
        for p in partners:
            if atypes[p][0] in ('N', 'O', 'S'):
                donors[p] = True
    return donors


def _getPosIonizable(mol):
    # arginine, lysine and histidine
    posIonizables = np.zeros(mol.numAtoms, dtype=bool)

    # ARG
    n_idxs = np.where(((mol.resname == 'ARG') | (mol.resname == 'AR0')) & (
        mol.atomtype == 'N') & (mol.name != 'N'))
    allc_idxs = np.where((mol.resname == 'ARG') & (
        mol.atomtype == 'C') & (mol.name != 'C'))[0]
    c_idxs = []
    for c in allc_idxs:
        bs = np.where(mol.bonds == c)[0]
        if len(bs) == 3:
            c_idxs.append(c)

    aidxs = n_idxs[0].tolist() + c_idxs

    # LYS
    n_idxs = np.where(((mol.resname == 'LYS') | (mol.resname == 'LYN')) & (
        mol.atomtype == 'N') & (mol.name != 'N'))
    aidxs += n_idxs[0].tolist()

    # HIS, HID, HIE, HIP, HSD, HSE
    n_idxs = np.where(((mol.resname == 'HIS') | (mol.resname == 'HID') |
                       (mol.resname == 'HIE') | (mol.resname == 'HIP') |
                       (mol.resname == 'HSE') | (mol.resname == 'HSD') |
                       (mol.resname == 'HSP')) &
                      ((mol.atomtype == 'N') | (mol.atomtype == 'NA') | (mol.atomtype == 'Nn') | (mol.atomtype == 'Na')) &
                      (mol.name != 'N'))

    c_idxs = np.where(((mol.resname == 'HIS') | (mol.resname == 'HID') |
                       (mol.resname == 'HIE') | (mol.resname == 'HIP') |
                       (mol.resname == 'HSE') | (mol.resname == 'HSD') |
                       (mol.resname == 'HSP')) &
                      (mol.atomtype == 'A'))

    aidxs += n_idxs[0].tolist() + c_idxs[0].tolist()

    posIonizables[aidxs] = 1

    return posIonizables


def _getNegIonizable(mol):
    # aspartic and glutamate
    negIonizables = np.zeros(mol.numAtoms, dtype=bool)

    # ASP
    o_idxs = np.where(((mol.resname == 'ASP') | (mol.resname == 'ASH')) &
                      (mol.atomtype == 'OA') & (mol.name != 'O'))
    allc_idxs = np.where(((mol.resname == 'ASP') | (mol.resname == 'ASH')) &
                         (mol.atomtype == 'C') & (mol.name != 'C'))[0]
    c_idxs = []
    for c in allc_idxs:
        bs = np.where(mol.bonds == c)[0]
        if len(bs) == 3:
            c_idxs.append(c)
    aidxs = o_idxs[0].tolist() + c_idxs

    # Glutamate
    o_idxs = np.where(((mol.resname == 'GLU') | (mol.resname == 'GLH')) &
                      (mol.atomtype == 'OA') & (mol.name != 'O'))

    allc_idxs = np.where(((mol.resname == 'GLU') | (mol.resname == 'GLH')) &
                         (mol.atomtype == 'C') & (mol.name != 'C'))[0]
    c_idxs = []
    for c in allc_idxs:
        bs = np.where(mol.bonds == c)[0]
        if len(bs) == 3:
            c_idxs.append(c)
    aidxs += o_idxs[0].tolist() + c_idxs

    negIonizables[aidxs] = 1

    return negIonizables


def _getOccupancy(elements):
    return np.array(elements) != 'H'


def _getMetals(atypes):
    return (atypes == 'MG') | (atypes == 'ZN') | (atypes == 'MN') | \
           (atypes == 'CA') | (atypes == 'FE') | (atypes == 'HG') | \
           (atypes == 'CD') | (atypes == 'NI') | (atypes == 'CO') | \
           (atypes == 'CU') | (atypes == 'K') | (atypes == 'LI') | \
           (atypes == 'Mg') | (atypes == 'Zn') | (atypes == 'Mn') | \
           (atypes == 'Ca') | (atypes == 'Fe') | (atypes == 'Hg') | \
           (atypes == 'Cd') | (atypes == 'Ni') | (atypes == 'Co') | \
           (atypes == 'Cu') | (atypes == 'Li')


def getFeatures(mol):
    atypes = mol.atomtype
    elements = [el[0] for el in atypes]

    hydr = _getHydrophobic(atypes)
    arom = _getAromatic(atypes)
    acc = _getAcceptor(atypes)
    don = _getDonors(atypes, mol.bonds)
    pos = _getPosIonizable(mol)
    neg = _getNegIonizable(mol)
    metals = _getMetals(atypes)
    occ = _getOccupancy(elements)

    return np.vstack((hydr, arom, acc, don, pos, neg, metals, occ)).T.copy()


def parallel(func, listobj, n_cpus=-1, *args):
    from tqdm import tqdm
    from joblib import Parallel, delayed
    results = Parallel(n_jobs=n_cpus)(delayed(func)(ob, *args)
                                      for ob in tqdm(listobj))
    return results


# (c) 2015-2018 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#
from moleculekit.projections.projection import Projection
import numpy as np
import logging
logger = logging.getLogger(__name__)


class MetricTMscore(Projection):
    """ Calculates the TMscore of a set of trajectories to a reference structure

    Parameters
    ----------
    refmol : :class:`Molecule <moleculekit.molecule.Molecule>` object
        The reference Molecule to which we want to calculate the TMscore.
    trajtmstr : str
        Atom selection string for the trajectories from which to calculate the TMscore.
        See more `here <http://www.ks.uiuc.edu/Research/vmd/vmd-1.9.2/ug/node89.html>`__
    reftmstr : str, optional
        Atom selection string for the reference structure from which to calculate the TMscore. If None, it defaults to
        `trajrmsdstr`. See more `here <http://www.ks.uiuc.edu/Research/vmd/vmd-1.9.2/ug/node89.html>`__
    centerstr : str, optional
        Atom selection string around which to center the wrapping of the trajectories.
        See more `here <http://www.ks.uiuc.edu/Research/vmd/vmd-1.9.2/ug/node89.html>`__
    """
    def __init__(self, refmol, trajtmstr, reftmstr=None):
        super().__init__()

        if reftmstr is None:
            reftmstr = trajtmstr
        self._refmol = refmol
        self._reftmsel = self._refmol.atomselect(reftmstr) & (self._refmol.name == 'CA')
        self._trajtmsel = trajtmstr

    def _calculateMolProp(self, mol, props='all'):
        res = {}
        res['trajtmsel'] = mol.atomselect(self._trajtmsel) & (mol.name == 'CA')
        if np.sum(res['trajtmsel']) == 0:
            raise NameError('RMSD atom selection resulted in 0 atoms.')
        return res

    def project(self, mol):
        """ Project molecule.

        Parameters
        ----------
        mol : :class:`Molecule <moleculekit.molecule.Molecule>`
            A :class:`Molecule <moleculekit.molecule.Molecule>` object to project.

        Returns
        -------
        data : np.ndarray
            An array containing the projected data.
        """
        from moleculekit.util import molTMscore
        mol = mol.copy()
        trajtmsel = self._getMolProp(mol, 'trajtmsel')

        tm, _ = molTMscore(mol, self._refmol, trajtmsel, self._reftmsel)
        return tm[:, np.newaxis]

    def getMapping(self, mol):
        """ Returns the description of each projected dimension.

        Parameters
        ----------
        mol : :class:`Molecule <moleculekit.molecule.Molecule>` object
            A Molecule object which will be used to calculate the descriptions of the projected dimensions.

        Returns
        -------
        map : :class:`DataFrame <pandas.core.frame.DataFrame>` object
            A DataFrame containing the descriptions of each dimension
        """
        trajtmsel = self._getMolProp(mol, 'trajtmsel')
        from pandas import DataFrame
        types = ['tmscore']
        indexes = [np.where(trajtmsel)[0]]
        description = ['TMscore to reference structure.']
        return DataFrame({'type': types, 'atomIndexes': indexes, 'description': description})


import unittest
class _TestMetricTMscore(unittest.TestCase):
    def test_tmscore(self):
        from moleculekit.molecule import Molecule
        from moleculekit.home import home
        import numpy as np
        from os import path

        mol = Molecule(path.join(home(dataDir='test-projections'), 'trajectory', 'filtered.pdb'))
        mol.read(path.join(home(dataDir='test-projections'), 'trajectory', 'traj.xtc'))
        ref = mol.copy()

        ref.dropFrames(keep=0)
        mol.dropFrames(keep=np.arange(mol.numFrames-20, mol.numFrames))

        metr = MetricTMscore(ref, 'protein and name CA')
        data = metr.project(mol)

        lasttm = np.array([0.9633381, 0.96441294, 0.96553609, 0.96088852, 0.96288511, 0.95677591, 0.96544727, 0.96359811,
                        0.95658912, 0.96893117, 0.96623924, 0.96064913, 0.96207041, 0.95947848, 0.96657048, 0.95993426,
                        0.96543296, 0.96806875, 0.96437248, 0.96144066], dtype=np.float32)
        assert np.all(np.abs(data.flatten() - lasttm) < 0.001), 'TMscore calculation is broken'


if __name__ == "__main__":
    unittest.main(verbosity=2)



#
# !!!! REMOVE THIS FILE !!!!
#
#
#
# import logging
# import tempfile
#
# import propka
# import propka.lib
# import propka.molecular_container
#
# from htmd.builder.preparationdata import PreparationData
#
# logger = logging.getLogger(__name__)
#
#
# def pka(mol, pH=7.0):
#     """Compute pKa values via propKa 3.1
#
#     Parameters
#     ----------
#     mol : Molecule
#         The molecule to be analyzed
#     pH : float
#         The pH at which to do the analysis
#
#     Returns
#     -------
#     rd : DataFrame
#         A pandas dataframe containing resid, chain, pKa arrays.
#         Properties unrelated to pKa are unset.
#
#     See Also
#     --------
#     The ResidueData object.
#
#
#     Examples
#     --------
#     >>> m=Molecule("3ptb")
#     >>> rd = pka(m)
#     >>> rd.pKa[rd.resid==189]
#     39    4.949079
#     Name: pKa, dtype: float64
#     """
#
#     return _pka_backend_internal(mol, pH)
#
#
#
# def _pka_backend_internal(mol, pH):
#     tmpmol = mol.copy()
#     tmpmol.filter("noh")
#     pka_pdb = tempfile.NamedTemporaryFile(mode="w+", suffix=".pdb")
#     tmpmol.write(pka_pdb.name)
#
#     pka_options, _ = propka.lib.loadOptions('--pH', pH, '--quiet')
#     pka_molecule = propka.molecular_container.Molecular_container(pka_pdb.name, pka_options)
#     pka_pdb.close()
#
#     # calculating pKa values for ionizable residues -
#     pka_molecule.calculate_pka()
#
#     rd = PreparationData()
#     rd._importPKAs(pka_molecule)
#     return rd.data
#
#
#
# def _pka_backend_external(molecule, propka=None, pH=7.0):
#     '''PKA - Predict per-residue pKa, using propka
#
#     Parameters
#     ----------
#     molecule : A Molecule object
#     propka   : Path of propka program. Default: find automatically
#     pH       : pH at which to perform calculation. Default: 7.0
#
#     Return
#     ------
#
#     List of dicts containin the keys 'resname', 'resid', 'chainid', 'pka'
#
#     Example
#     -------
#
#     pkas = pka( Molecule('4DFR'), pH=6.5 )
#     '''
#
#     from tempfile import mkstemp
#     import os
#     from shutil import which
#     from subprocess import check_output
#
#     try:
#         propka = which("propka31", mode=os.X_OK)
#     except:
#         raise NameError("Cannot find 'propka31' in PATH");
#
#     if not os.access(propka, os.X_OK):
#         raise NameError("'propka' not found");
#
#     try:
#         pH = float(pH)
#     except:
#         raise NameError("pH value invalid")
#     if (pH <= 0. or pH >= 14.):
#         raise NameError("pH value invalid")
#
#     fh, fnp = mkstemp()
#     fn = fnp + ".pdb"
#     molecule.write(fn)
#
#     dd = os.getcwd()
#     try:
#         os.chdir(os.path.dirname(fn))
#         op = check_output([propka, fn, "-o", str(pH)])
#         op = op.decode("ascii").split("\n");
#     except:
#         raise NameError("Failed to execute PropKa")
#     os.chdir(dd)
#
#     ret = []
#     # print(op)
#     atpka = False
#     for line in op:
#         if not atpka:
#             if "model-pKa" in line:
#                 atpka = True
#                 continue
#             else:
#                 continue
#             if "-" in line:
#                 atpka = False
#                 continue
#         print(line)
#         s = line.split()
#         if len(s) >= 4:
#             pka = {"resname": s[0], "resid": s[1], "chainid": s[2], "pka": float(s[3])}
#             ret.append(pka)
#     os.unlink(fn)
#     os.unlink(fnp + ".pka")
#     os.unlink(fnp + ".propka_input")
#     return ret
#
#
#
#
# # A test method
# if __name__ == "__main__":
#     import doctest
#     from moleculekit.molecule import Molecule
#     doctest.testmod()
#
